#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import datetime
from collections import OrderedDict
from dataclasses import dataclass
from enum import Enum
from functools import partial
from inspect import isclass
from io import StringIO
from logging import Logger
from typing import Any

import numpy as np
import pandas as pd
import torch
from ax.adapter.registry import _decode_callables_from_references, GeneratorRegistryBase
from ax.core.base_trial import BaseTrial
from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.core.generator_run import GeneratorRun
from ax.core.multi_type_experiment import MultiTypeExperiment
from ax.core.objective import Objective
from ax.core.parameter import Parameter
from ax.core.parameter_constraint import (
    OrderConstraint,
    ParameterConstraint,
    SumConstraint,
)
from ax.core.search_space import SearchSpace
from ax.exceptions.storage import JSONDecodeError, STORAGE_DOCS_SUFFIX
from ax.generation_strategy.generation_node_input_constructors import (
    InputConstructorPurpose,
)
from ax.generation_strategy.generation_strategy import (
    GenerationNode,
    GenerationStep,
    GenerationStrategy,
)
from ax.generation_strategy.generator_spec import GeneratorSpec
from ax.generation_strategy.transition_criterion import (
    AuxiliaryExperimentCheck,
    TransitionCriterion,
    TrialBasedCriterion,
)
from ax.generators.torch.botorch_modular.generator import BoTorchGenerator
from ax.generators.torch.botorch_modular.surrogate import Surrogate, SurrogateSpec
from ax.generators.torch.botorch_modular.utils import ModelConfig
from ax.storage.json_store.decoders import (
    batch_trial_from_json,
    botorch_component_from_json,
    tensor_from_json,
    trial_from_json,
)
from ax.storage.json_store.registry import (
    CORE_CLASS_DECODER_REGISTRY,
    CORE_DECODER_REGISTRY,
)
from ax.utils.common.logger import get_logger
from ax.utils.common.serialization import (
    SerializationMixin,
    TClassDecoderRegistry,
    TDecoderRegistry,
)
from ax.utils.common.typeutils_torch import torch_type_from_str
from botorch.utils.types import DEFAULT
from pyre_extensions import assert_is_instance, none_throws


logger: Logger = get_logger(__name__)

# Deprecated model registry entries and their replacements.
# Used below in `_update_deprecated_model_registry`.
_DEPRECATED_MODEL_TO_REPLACEMENT: dict[str, str] = {
    "GPEI": "BOTORCH_MODULAR",
    "MOO": "BOTORCH_MODULAR",
    "FULLYBAYESIAN": "SAASBO",
    "FULLYBAYESIANMOO": "SAASBO",
    "FULLYBAYESIAN_MTGP": "SAAS_MTGP",
    "FULLYBAYESIANMOO_MTGP": "SAAS_MTGP",
    "ST_MTGP_LEGACY": "ST_MTGP",
    "ST_MTGP_NEHVI": "ST_MTGP",
}

# Deprecated model kwargs, to be removed from GStep / GNodes.
_DEPRECATED_MODEL_KWARGS: tuple[str, ...] = (
    "fit_on_update",
    "torch_dtype",
    "status_quo_name",
    "status_quo_features",
)

# Deprecated node input constructors, removed from GNodes.
# NOTE: These are the enum keys, which are typically upper-case.
_DEPRECATED_NODE_INPUT_CONSTRUCTORS: tuple[str, ...] = ("STATUS_QUO_FEATURES",)


@dataclass
class RegistryKwargs:
    decoder_registry: TDecoderRegistry
    class_decoder_registry: TClassDecoderRegistry


def object_from_json(
    object_json: Any,
    decoder_registry: TDecoderRegistry = CORE_DECODER_REGISTRY,
    class_decoder_registry: TClassDecoderRegistry = CORE_CLASS_DECODER_REGISTRY,
) -> Any:
    """Recursively load objects from a JSON-serializable dictionary."""

    registry_kwargs = RegistryKwargs(
        decoder_registry=decoder_registry, class_decoder_registry=class_decoder_registry
    )

    _object_from_json = partial(object_from_json, **vars(registry_kwargs))

    if type(object_json) in (str, int, float, bool, type(None)) or isinstance(
        object_json, Enum
    ):
        return object_json
    elif isinstance(object_json, list):
        return [_object_from_json(i) for i in object_json]
    elif isinstance(object_json, tuple):
        return tuple(_object_from_json(i) for i in object_json)
    elif isinstance(object_json, dict):
        if "__type" not in object_json:
            # this is just a regular dictionary, e.g. the one in Parameter
            # containing parameterizations
            return {k: _object_from_json(v) for k, v in object_json.items()}

        _type = object_json.pop("__type")

        if _type == "datetime":
            return datetime.datetime.strptime(
                object_json["value"], "%Y-%m-%d %H:%M:%S.%f"
            )
        elif _type == "OrderedDict":
            return OrderedDict(
                [(k, _object_from_json(v)) for k, v in object_json["value"]]
            )
        elif _type == "DataFrame":
            # Need dtype=False, otherwise infers arm_names like "4_1"
            # should be int 41
            return pd.read_json(StringIO(object_json["value"]), dtype=False)
        elif _type == "ndarray":
            return np.array(object_json["value"])
        elif _type == "Tensor":
            return tensor_from_json(json=object_json)
        elif _type.startswith("torch"):
            # Torch types will be encoded as "torch_<type_name>", so we drop prefix
            return torch_type_from_str(
                identifier=object_json["value"], type_name=_type[6:]
            )
        elif _type == "ListSurrogate":
            return surrogate_from_list_surrogate_json(
                list_surrogate_json=object_json, **vars(registry_kwargs)
            )
        elif _type == "set":
            return set(object_json["value"])
        # Used for decoding classes (not objects).
        elif _type in class_decoder_registry:
            return class_decoder_registry[_type](object_json)

        elif _type == "GeneratorRunStruct":
            object_json.pop("weight", None)  # Deprecated.

        elif _type not in decoder_registry:
            err = (
                f"The JSON dictionary passed to `object_from_json` has a type "
                f"{_type} that is not registered with a corresponding class in "
                f"DECODER_REGISTRY. {STORAGE_DOCS_SUFFIX}"
            )
            raise JSONDecodeError(err)

        # pyre-fixme[9, 24]: Generic type `type` expects 1 type parameter, use
        # `typing.Type[<base type>]` to avoid runtime subscripting errors.
        _class: type = decoder_registry[_type]
        if isclass(_class) and issubclass(_class, Enum):
            name = object_json["name"]
            if issubclass(_class, GeneratorRegistryBase):
                name = _update_deprecated_model_registry(name=name)
            # to access enum members by name, use item access
            return _class[name]
        elif isclass(_class) and issubclass(_class, torch.nn.Module):
            return botorch_component_from_json(botorch_class=_class, json=object_json)
        elif _class == GeneratorRun:
            return generator_run_from_json(
                object_json=object_json, **vars(registry_kwargs)
            )
        elif _class == GenerationStep:
            return generation_step_from_json(
                generation_step_json=object_json, **vars(registry_kwargs)
            )
        elif _class == GenerationNode:
            return generation_node_from_json(
                generation_node_json=object_json, **vars(registry_kwargs)
            )
        elif _class == GeneratorSpec:
            return generator_spec_from_json(
                generator_spec_json=object_json, **vars(registry_kwargs)
            )
        elif _class == GenerationStrategy:
            return generation_strategy_from_json(
                generation_strategy_json=object_json, **vars(registry_kwargs)
            )
        elif _class == MultiTypeExperiment:
            return multi_type_experiment_from_json(
                object_json=object_json, **vars(registry_kwargs)
            )
        elif _class == Experiment:
            return experiment_from_json(
                object_json=object_json, **vars(registry_kwargs)
            )
        elif _class == SearchSpace:
            return search_space_from_json(
                search_space_json=object_json, **vars(registry_kwargs)
            )
        elif _class == Objective:
            return objective_from_json(object_json=object_json, **vars(registry_kwargs))
        elif _class in (SurrogateSpec, Surrogate, ModelConfig):
            if "input_transform" in object_json:
                (
                    input_transform_classes_json,
                    input_transform_options_json,
                ) = get_input_transform_json_components(
                    input_transforms_json=object_json.pop("input_transform"),
                    **vars(registry_kwargs),
                )
                object_json["input_transform_classes"] = input_transform_classes_json
                object_json["input_transform_options"] = input_transform_options_json
            if "outcome_transform" in object_json:
                (
                    outcome_transform_classes_json,
                    outcome_transform_options_json,
                ) = get_outcome_transform_json_components(
                    outcome_transforms_json=object_json.pop("outcome_transform"),
                    **vars(registry_kwargs),
                )
                object_json["outcome_transform_classes"] = (
                    outcome_transform_classes_json
                )
                object_json["outcome_transform_options"] = (
                    outcome_transform_options_json
                )
        elif isclass(_class) and (
            issubclass(_class, TrialBasedCriterion)
            or issubclass(_class, AuxiliaryExperimentCheck)
        ):
            # TrialBasedCriterion contains a list of `TrialStatus` for args.
            # AuxiliaryExperimentCheck contains AuxiliaryExperimentPurpose objects
            # They need to be unpacked by hand to properly retain the types.
            return unpack_transition_criteria_from_json(
                class_=_class,
                transition_criteria_json=object_json,
                **vars(registry_kwargs),
            )
        elif isclass(_class) and issubclass(_class, SerializationMixin):
            return _class(
                # Note: we do not recursively call object_from_json here again as
                # that would invalidate design principles behind deserialize_init_args.
                # Any Ax class that needs serialization and who's init args include
                # another Ax class that needs serialization should implement its own
                # _to_json and _from_json methods and register them appropriately.
                **_class.deserialize_init_args(
                    args=object_json, **vars(registry_kwargs)
                )
            )
        if _class in (BoTorchGenerator, Surrogate):
            # Updates deprecated surrogate spec related inputs.
            object_json = _sanitize_surrogate_spec_input(object_json=object_json)
        if _class is Surrogate:
            object_json = _sanitize_legacy_surrogate_inputs(object_json=object_json)
        if _class is SurrogateSpec:
            object_json = _sanitize_inputs_to_surrogate_spec(object_json=object_json)

        return ax_class_from_json_dict(
            _class=_class, object_json=object_json, **vars(registry_kwargs)
        )
    else:
        err = (
            f"The object {object_json} passed to `object_from_json` has an "
            f"unsupported type: {type(object_json)}."
        )
        raise JSONDecodeError(err)


def ax_class_from_json_dict(
    # pyre-fixme[24]: Generic type `type` expects 1 type parameter, use
    #  `typing.Type` to avoid runtime subscripting errors.
    _class: type,
    object_json: dict[str, Any],
    decoder_registry: TDecoderRegistry = CORE_DECODER_REGISTRY,
    class_decoder_registry: TClassDecoderRegistry = CORE_CLASS_DECODER_REGISTRY,
) -> Any:
    """Reinstantiates an Ax class registered in `DECODER_REGISTRY` from a JSON
    dict.
    """
    return _class(
        **{
            k: object_from_json(
                v,
                decoder_registry=decoder_registry,
                class_decoder_registry=class_decoder_registry,
            )
            for k, v in object_json.items()
        }
    )


def generator_run_from_json(
    object_json: dict[str, Any],
    decoder_registry: TDecoderRegistry = CORE_DECODER_REGISTRY,
    class_decoder_registry: TClassDecoderRegistry = CORE_CLASS_DECODER_REGISTRY,
) -> GeneratorRun:
    """Load Ax GeneratorRun from JSON."""
    time_created_json = object_json.pop("time_created")
    type_json = object_json.pop("generator_run_type")
    object_json.pop("index", None)  # Deprecated.
    # Remove `objective_thresholds` to avoid issues with registries, since
    # `ObjectiveThreshold` depend on `Metric` objects.
    object_json.pop("objective_thresholds", None)
    generator_run = GeneratorRun(
        **{
            k: object_from_json(
                v,
                decoder_registry=decoder_registry,
                class_decoder_registry=class_decoder_registry,
            )
            for k, v in object_json.items()
        }
    )
    # NOTE: JSON converts all tuples to lists, and we need to convert them back. This is
    # an ad hoc fix, though.
    if isinstance(generator_run._best_arm_predictions, list):
        arm, arm_prediction = generator_run._best_arm_predictions
        if arm_prediction is not None and isinstance(arm_prediction, list):
            arm_prediction = tuple(arm_prediction)
        generator_run._best_arm_predictions = (arm, arm_prediction)

    if isinstance(generator_run._model_predictions, list):
        generator_run._model_predictions = tuple(generator_run._model_predictions)

    # Remove deprecated kwargs from model kwargs & adapter kwargs.
    if generator_run._model_kwargs is not None:
        generator_run._model_kwargs = {
            k: v
            for k, v in generator_run._model_kwargs.items()
            if k not in _DEPRECATED_MODEL_KWARGS
        }
    if generator_run._bridge_kwargs is not None:
        generator_run._bridge_kwargs = {
            k: v
            for k, v in generator_run._bridge_kwargs.items()
            if k not in _DEPRECATED_MODEL_KWARGS
        }
    generator_run._time_created = object_from_json(
        time_created_json,
        decoder_registry=decoder_registry,
        class_decoder_registry=class_decoder_registry,
    )
    generator_run._generator_run_type = object_from_json(
        type_json,
        decoder_registry=decoder_registry,
        class_decoder_registry=class_decoder_registry,
    )
    return generator_run


def unpack_transition_criteria_from_json(
    # pyre-fixme[24]: Generic type `type` expects 1 type parameter, use `typing.Type` to
    #  avoid runtime subscripting errors.
    class_: type,
    transition_criteria_json: dict[str, Any],
    decoder_registry: TDecoderRegistry = CORE_DECODER_REGISTRY,
    class_decoder_registry: TClassDecoderRegistry = CORE_CLASS_DECODER_REGISTRY,
) -> TransitionCriterion | None:
    """Load Ax transition criteria that depend on Trials from JSON.

    Since ``TrialBasedCriterion`` contain lists of ``TrialStatus``,
    the json for these criterion needs to be carefully unpacked and
    re-processed via ``object_from_json`` in order to maintain correct
    typing. We pass in ``class_`` in order to correctly handle all classes
    which inherit from ``TrialBasedCriterion`` (ex: ``MaxTrials``).
    """
    new_dict = {}
    for key, value in transition_criteria_json.items():
        new_val = object_from_json(
            object_json=value,
            decoder_registry=decoder_registry,
            class_decoder_registry=class_decoder_registry,
        )
        new_dict[key] = new_val

    return class_(**new_dict)


def search_space_from_json(
    search_space_json: dict[str, Any],
    decoder_registry: TDecoderRegistry = CORE_DECODER_REGISTRY,
    class_decoder_registry: TClassDecoderRegistry = CORE_CLASS_DECODER_REGISTRY,
) -> SearchSpace:
    """Load a SearchSpace from JSON.

    This function is necessary due to the coupled loading of SearchSpace
    and parameter constraints.
    """
    parameters = object_from_json(
        search_space_json.pop("parameters"),
        decoder_registry=decoder_registry,
        class_decoder_registry=class_decoder_registry,
    )
    json_param_constraints = search_space_json.pop("parameter_constraints")
    return SearchSpace(
        parameters=parameters,
        parameter_constraints=parameter_constraints_from_json(
            parameter_constraint_json=json_param_constraints,
            parameters=parameters,
            decoder_registry=decoder_registry,
            class_decoder_registry=class_decoder_registry,
        ),
    )


def parameter_constraints_from_json(
    parameter_constraint_json: list[dict[str, Any]],
    parameters: list[Parameter],
    decoder_registry: TDecoderRegistry = CORE_DECODER_REGISTRY,
    class_decoder_registry: TClassDecoderRegistry = CORE_CLASS_DECODER_REGISTRY,
) -> list[ParameterConstraint]:
    """Load ParameterConstraints from JSON.

    Order and SumConstraint are tied to a search space,
    and require that SearchSpace's parameters to be passed in for decoding.

    Args:
        parameter_constraint_json: JSON representation of parameter constraints.
        parameters: Parameter definitions for decoding via parameter names.

    Returns:
        parameter_constraints: Python classes for parameter constraints.
    """
    parameter_constraints = []
    parameter_map = {p.name: p for p in parameters}
    for constraint in parameter_constraint_json:
        if constraint["__type"] == "OrderConstraint":
            lower_parameter = parameter_map[constraint["lower_name"]]
            upper_parameter = parameter_map[constraint["upper_name"]]
            parameter_constraints.append(
                OrderConstraint(
                    lower_parameter=lower_parameter, upper_parameter=upper_parameter
                )
            )
        elif constraint["__type"] == "SumConstraint":
            parameters = [parameter_map[name] for name in constraint["parameter_names"]]
            parameter_constraints.append(
                SumConstraint(
                    parameters=parameters,
                    is_upper_bound=constraint["is_upper_bound"],
                    bound=constraint["bound"],
                )
            )
        else:
            parameter_constraints.append(
                object_from_json(
                    constraint,
                    decoder_registry=decoder_registry,
                    class_decoder_registry=class_decoder_registry,
                )
            )
    return parameter_constraints


def trials_from_json(
    experiment: Experiment,
    trials_json: dict[str, Any],
    decoder_registry: TDecoderRegistry = CORE_DECODER_REGISTRY,
    class_decoder_registry: TClassDecoderRegistry = CORE_CLASS_DECODER_REGISTRY,
) -> dict[int, BaseTrial]:
    """Load Ax Trials from JSON."""
    loaded_trials = {}
    for index, batch_json in trials_json.items():
        is_trial = batch_json["__type"] == "Trial"
        batch_json = {
            k: object_from_json(
                v,
                decoder_registry=decoder_registry,
                class_decoder_registry=class_decoder_registry,
            )
            for k, v in batch_json.items()
            if k != "__type"
        }
        loaded_trials[int(index)] = (
            trial_from_json(experiment=experiment, **batch_json)
            if is_trial
            else batch_trial_from_json(experiment=experiment, **batch_json)
        )
    return loaded_trials


def data_from_json(
    data_by_trial_json: dict[str, Any],
    decoder_registry: TDecoderRegistry = CORE_DECODER_REGISTRY,
    class_decoder_registry: TClassDecoderRegistry = CORE_CLASS_DECODER_REGISTRY,
) -> dict[int, "OrderedDict[int, Data]"]:
    """Load Ax Data from JSON."""
    data_by_trial = object_from_json(
        data_by_trial_json,
        decoder_registry=decoder_registry,
        class_decoder_registry=class_decoder_registry,
    )
    # hack necessary because Python's json module converts dictionary
    # keys to strings: https://stackoverflow.com/q/1450957
    return {
        int(k): OrderedDict({int(k2): v2 for k2, v2 in v.items()})
        for k, v in data_by_trial.items()
    }


def multi_type_experiment_from_json(
    object_json: dict[str, Any],
    decoder_registry: TDecoderRegistry = CORE_DECODER_REGISTRY,
    class_decoder_registry: TClassDecoderRegistry = CORE_CLASS_DECODER_REGISTRY,
) -> MultiTypeExperiment:
    """Load AE MultiTypeExperiment from JSON."""
    experiment_info = _get_experiment_info(object_json)

    _metric_to_canonical_name = object_json.pop("_metric_to_canonical_name")
    _metric_to_trial_type = object_json.pop("_metric_to_trial_type")
    _trial_type_to_runner = object_from_json(
        object_json.pop("_trial_type_to_runner"),
        decoder_registry=decoder_registry,
        class_decoder_registry=class_decoder_registry,
    )
    tracking_metrics = object_from_json(
        object_json.pop("tracking_metrics"),
        decoder_registry=decoder_registry,
        class_decoder_registry=class_decoder_registry,
    )
    # not relevant to multi type experiment
    del object_json["runner"]

    kwargs = {
        k: object_from_json(
            v,
            decoder_registry=decoder_registry,
            class_decoder_registry=class_decoder_registry,
        )
        for k, v in object_json.items()
    }
    kwargs["default_runner"] = _trial_type_to_runner[object_json["default_trial_type"]]

    experiment = MultiTypeExperiment(**kwargs)
    for metric in tracking_metrics:
        experiment._tracking_metrics[metric.name] = metric
    experiment._metric_to_canonical_name = _metric_to_canonical_name
    experiment._metric_to_trial_type = _metric_to_trial_type
    experiment._trial_type_to_runner = _trial_type_to_runner

    _load_experiment_info(
        exp=experiment,
        exp_info=experiment_info,
        decoder_registry=decoder_registry,
        class_decoder_registry=class_decoder_registry,
    )
    return experiment


def experiment_from_json(
    object_json: dict[str, Any],
    decoder_registry: TDecoderRegistry = CORE_DECODER_REGISTRY,
    class_decoder_registry: TClassDecoderRegistry = CORE_CLASS_DECODER_REGISTRY,
) -> Experiment:
    """Load Ax Experiment from JSON."""
    experiment_info = _get_experiment_info(object_json)

    experiment = Experiment(
        **{
            k: object_from_json(
                v,
                decoder_registry=decoder_registry,
                class_decoder_registry=class_decoder_registry,
            )
            for k, v in object_json.items()
        }
    )
    experiment._arms_by_name = {}

    _load_experiment_info(
        exp=experiment,
        exp_info=experiment_info,
        decoder_registry=decoder_registry,
        class_decoder_registry=class_decoder_registry,
    )
    return experiment


def _get_experiment_info(object_json: dict[str, Any]) -> dict[str, Any]:
    """Returns basic information from `Experiment` object_json."""
    return {
        "time_created_json": object_json.pop("time_created"),
        "trials_json": object_json.pop("trials"),
        "experiment_type_json": object_json.pop("experiment_type"),
        "data_by_trial_json": object_json.pop("data_by_trial"),
    }


def _load_experiment_info(
    exp: Experiment,
    exp_info: dict[str, Any],
    decoder_registry: TDecoderRegistry = CORE_DECODER_REGISTRY,
    class_decoder_registry: TClassDecoderRegistry = CORE_CLASS_DECODER_REGISTRY,
) -> None:
    """Loads `Experiment` object with basic information."""
    exp._time_created = object_from_json(
        exp_info.get("time_created_json"),
        decoder_registry=decoder_registry,
        class_decoder_registry=class_decoder_registry,
    )
    exp._trials = trials_from_json(
        exp,
        exp_info.get("trials_json"),
        decoder_registry=decoder_registry,
        class_decoder_registry=class_decoder_registry,
    )
    exp._experiment_type = object_from_json(
        exp_info.get("experiment_type_json"),
        decoder_registry=decoder_registry,
        class_decoder_registry=class_decoder_registry,
    )
    exp._data_by_trial = data_from_json(
        exp_info.get("data_by_trial_json"),
        decoder_registry=decoder_registry,
        class_decoder_registry=class_decoder_registry,
    )
    for trial in exp._trials.values():
        for arm in trial.arms:
            exp._register_arm(arm)
        if trial.ttl_seconds is not None:
            exp._trials_have_ttl = True
    if exp.status_quo is not None:
        sq = none_throws(exp.status_quo)
        exp._register_arm(sq)


def _convert_generation_step_keys_for_backwards_compatibility(
    object_json: dict[str, Any],
) -> dict[str, Any]:
    """If necessary, converts keys in a JSON dict representing a `GenerationStep`
    for backwards compatibility.
    """
    # NOTE: this is a hack to make generation steps able to load after the
    # renaming of generation step fields to be in terms of 'trials' rather than
    # 'arms'.
    keys = list(object_json.keys())
    for k in keys:
        if "arms" in k:
            object_json[k.replace("arms", "trials")] = object_json.pop(k)
        if k == "recommended_max_parallelism":
            object_json["max_parallelism"] = object_json.pop(k)
    return object_json


def generation_node_from_json(
    generation_node_json: dict[str, Any],
    decoder_registry: TDecoderRegistry = CORE_DECODER_REGISTRY,
    class_decoder_registry: TClassDecoderRegistry = CORE_CLASS_DECODER_REGISTRY,
) -> GenerationNode:
    """Load GenerationNode object from JSON."""
    # Due to input_constructors being a dictionary with both keys and values being of
    # type enum, we must manually decode them here because object_from_json doesn't
    # recursively decode dictionary key values.
    decoded_input_constructors = None
    if "input_constructors" in generation_node_json.keys():
        decoded_input_constructors = {}
        for key, value in generation_node_json.pop("input_constructors").items():
            if key in _DEPRECATED_NODE_INPUT_CONSTRUCTORS:
                # Skip deprecated input constructors.
                continue
            decoded_input_constructors[InputConstructorPurpose[key]] = object_from_json(
                value,
                decoder_registry=decoder_registry,
                class_decoder_registry=class_decoder_registry,
            )
    if "model_specs" in generation_node_json:
        # Check for all kwarg to support backwards compatibility.
        generator_specs = generation_node_json.pop("model_specs")
    else:
        generator_specs = generation_node_json.pop("generator_specs")
    return GenerationNode(
        node_name=generation_node_json.pop("node_name"),
        generator_specs=object_from_json(
            object_json=generator_specs,
            decoder_registry=decoder_registry,
            class_decoder_registry=class_decoder_registry,
        ),
        best_model_selector=object_from_json(
            object_json=generation_node_json.pop("best_model_selector", None),
            decoder_registry=decoder_registry,
            class_decoder_registry=class_decoder_registry,
        ),
        should_deduplicate=generation_node_json.pop("should_deduplicate", False),
        transition_criteria=(
            object_from_json(
                object_json=generation_node_json.pop("transition_criteria"),
                decoder_registry=decoder_registry,
                class_decoder_registry=class_decoder_registry,
            )
            if "transition_criteria" in generation_node_json.keys()
            else None
        ),
        input_constructors=decoded_input_constructors,
        previous_node_name=(
            generation_node_json.pop("previous_node_name")
            if "previous_node_name" in generation_node_json.keys()
            else None
        ),
        trial_type=(
            object_from_json(
                object_json=generation_node_json.pop("trial_type"),
                decoder_registry=decoder_registry,
                class_decoder_registry=class_decoder_registry,
            )
            if "trial_type" in generation_node_json.keys()
            else None
        ),
    )


def _sanitize_inputs_to_surrogate_spec(
    object_json: dict[str, Any],
) -> dict[str, Any]:
    """This is a backwards compatibility helper for inputs to ``SurrogateSpec``.
    It replaces the legacy inputs in the json with a  ``ModelConfig`` and discards
    the legacy inputs.
    """
    new_json = object_json.copy()
    # If no model configs are available, this spec was constructed using legacy inputs.
    # We will replace it with a model config constructed from the legacy inputs.
    # It is possible that both inputs are available, in which case we will discard the
    # legacy inputs and only keep the existing model config.
    model_configs = new_json.get("model_configs", [])
    new_config = [
        {
            "__type": "ModelConfig",
            "botorch_model_class": new_json.pop("botorch_model_class", None),
            "model_options": new_json.pop("botorch_model_kwargs", {}),
            "mll_class": new_json.pop("mll_class", None),
            "mll_options": new_json.pop("mll_kwargs", {}),
            "input_transform_classes": new_json.pop("input_transform_classes", DEFAULT),
            "input_transform_options": new_json.pop("input_transform_options", {})
            or {},  # Old default was None.
            "outcome_transform_classes": new_json.pop(
                "outcome_transform_classes", None
            ),
            "outcome_transform_options": new_json.pop("outcome_transform_options", {})
            or {},  # Old default was None.
            "covar_module_class": new_json.pop("covar_module_class", None),
            "covar_module_options": new_json.pop("covar_module_kwargs", {}) or {},
            "likelihood_class": new_json.pop("likelihood_class", None),
            "likelihood_options": new_json.pop("likelihood_kwargs", {}) or {},
            "name": "from deprecated args",
        },
    ]
    if len(model_configs) == 0:
        new_json["model_configs"] = new_config
    return new_json


def _sanitize_surrogate_spec_input(
    object_json: dict[str, Any],
) -> dict[str, Any]:
    """This is a backwards compatibility helper for ``SurrogateSpec`` related
    inputs to ``BoTorchGenerator``.

    If ``object_json`` includes a ``surrogate_specs`` key that is a dict
    with a single element, this method replaces it with `surrogate_spec`
    key with the value of that element.

    If the legacy inputs were used to initialize the ``SurrogateSpec``,
    a ``ModelConfig`` is constructed with the legacy inputs and the legacy
    inputs are discarded.

    Args:
        object_json: A dictionary of json encoded inputs to update.

    Returns:
        The json with the surrogate spec related inputs updated.
        If there are multiple elements in ``surrogate_specs``, the input is discarded
        after logging an exception. The default ``Surrogate`` will be used. Otherwise,
        returns a new dictionary with the ``surrogate_specs`` element replaced with
        ``surrogate_spec`` and legacy inputs replaced with ``ModelConfig``.
    """
    new_json = object_json.copy()
    specs = new_json.pop("surrogate_specs", None)
    if specs is None:
        return new_json
    if len(specs) > 1:
        logger.exception(
            "The input includes `surrogate_specs` with multiple elements. "
            "Support for multiple surrogates has been deprecated. "
            "Discarding the `surrogate_specs` input to facilitate loading "
            "of the experiment. The loaded object will utilize the default "
            "`Surrogate` and may not behave as expected."
        )
        return new_json

    spec = next(iter(specs.values()))
    new_json["surrogate_spec"] = _sanitize_inputs_to_surrogate_spec(object_json=spec)
    return new_json


def _sanitize_legacy_surrogate_inputs(
    object_json: dict[str, Any],
) -> dict[str, Any]:
    """This is a backwards compatibility helper for ``Surrogate`` that replaces
    the legacy top level inputs with a ``SurrogateSpec`` with a single ``ModelConfig``.
    """
    new_json = object_json.copy()
    if new_json.get("surrogate_spec", None) is None:
        config_json = {
            "__type": "ModelConfig",
            "botorch_model_class": new_json.pop("botorch_model_class", None),
            "model_options": new_json.pop("model_options", {}),
            "mll_class": new_json.pop("mll_class", None),
            "mll_options": new_json.pop("mll_options", {}),
            "input_transform_classes": new_json.pop("input_transform_classes", DEFAULT),
            "input_transform_options": new_json.pop("input_transform_options", {})
            or {},  # Old default was None.
            "outcome_transform_classes": new_json.pop(
                "outcome_transform_classes", None
            ),
            "outcome_transform_options": new_json.pop("outcome_transform_options", {})
            or {},  # Old default was None.
            "covar_module_class": new_json.pop("covar_module_class", None),
            "covar_module_options": new_json.pop("covar_module_options", {}) or {},
            "likelihood_class": new_json.pop("likelihood_class", None),
            "likelihood_options": new_json.pop("likelihood_options", {}) or {},
            "name": "from deprecated args",
        }
        new_json["surrogate_spec"] = {
            "__type": "SurrogateSpec",
            "model_configs": [config_json],
            "allow_batched_models": new_json.pop("allow_batched_models", True),
        }
    return new_json


def generation_step_from_json(
    generation_step_json: dict[str, Any],
    decoder_registry: TDecoderRegistry = CORE_DECODER_REGISTRY,
    class_decoder_registry: TClassDecoderRegistry = CORE_CLASS_DECODER_REGISTRY,
) -> GenerationStep:
    """Load generation step from JSON."""
    generation_step_json = _convert_generation_step_keys_for_backwards_compatibility(
        generation_step_json
    )
    kwargs = generation_step_json.pop("model_kwargs", None)
    for k in _DEPRECATED_MODEL_KWARGS:
        # Remove deprecated kwargs.
        kwargs.pop(k, None)
    if kwargs is not None:
        kwargs = _sanitize_surrogate_spec_input(object_json=kwargs)
    gen_kwargs = generation_step_json.pop("model_gen_kwargs", None)
    completion_criteria = (
        object_from_json(
            generation_step_json.pop("completion_criteria"),
            decoder_registry=decoder_registry,
            class_decoder_registry=class_decoder_registry,
        )
        if "completion_criteria" in generation_step_json.keys()
        else []
    )
    if "model" in generation_step_json:
        # Old arg name for backwards compatibility.
        generator_json = generation_step_json.pop("model")
    else:
        generator_json = generation_step_json.pop("generator")
    generation_step = GenerationStep(
        generator=object_from_json(
            object_json=generator_json,
            decoder_registry=decoder_registry,
            class_decoder_registry=class_decoder_registry,
        ),
        num_trials=generation_step_json.pop("num_trials"),
        min_trials_observed=generation_step_json.pop("min_trials_observed", 0),
        completion_criteria=(
            completion_criteria if completion_criteria is not None else []
        ),
        max_parallelism=(generation_step_json.pop("max_parallelism", None)),
        enforce_num_trials=generation_step_json.pop("enforce_num_trials", True),
        model_kwargs=(
            _decode_callables_from_references(
                object_from_json(
                    kwargs,
                    decoder_registry=decoder_registry,
                    class_decoder_registry=class_decoder_registry,
                ),
            )
            if kwargs
            else {}
        ),
        model_gen_kwargs=(
            _decode_callables_from_references(
                object_from_json(
                    gen_kwargs,
                    decoder_registry=decoder_registry,
                    class_decoder_registry=class_decoder_registry,
                ),
            )
            if gen_kwargs
            else {}
        ),
        index=generation_step_json.pop("index", -1),
        should_deduplicate=generation_step_json.pop("should_deduplicate", False),
        generator_name=generation_step_json.pop("generator_name", None),
    )
    return generation_step


def generator_spec_from_json(
    generator_spec_json: dict[str, Any],
    decoder_registry: TDecoderRegistry = CORE_DECODER_REGISTRY,
    class_decoder_registry: TClassDecoderRegistry = CORE_CLASS_DECODER_REGISTRY,
) -> GeneratorSpec:
    """Load GeneratorSpec from JSON."""
    kwargs = generator_spec_json.pop("model_kwargs", None)
    for k in _DEPRECATED_MODEL_KWARGS:
        # Remove deprecated model kwargs.
        kwargs.pop(k, None)
    if kwargs is not None:
        kwargs = _sanitize_surrogate_spec_input(object_json=kwargs)
    gen_kwargs = generator_spec_json.pop("model_gen_kwargs", None)
    if "model_enum" in generator_spec_json:
        # Old arg name for backwards compatibility.
        enum = generator_spec_json.pop("model_enum")
    else:
        enum = generator_spec_json.pop("generator_enum")
    return GeneratorSpec(
        generator_enum=object_from_json(
            object_json=enum,
            decoder_registry=decoder_registry,
            class_decoder_registry=class_decoder_registry,
        ),
        model_kwargs=(
            _decode_callables_from_references(
                object_from_json(
                    object_json=kwargs,
                    decoder_registry=decoder_registry,
                    class_decoder_registry=class_decoder_registry,
                ),
            )
            if kwargs
            else {}
        ),
        model_gen_kwargs=(
            _decode_callables_from_references(
                object_from_json(
                    object_json=gen_kwargs,
                    decoder_registry=decoder_registry,
                    class_decoder_registry=class_decoder_registry,
                ),
            )
            if gen_kwargs
            else {}
        ),
    )


def generation_strategy_from_json(
    generation_strategy_json: dict[str, Any],
    experiment: Experiment | None = None,
    decoder_registry: TDecoderRegistry = CORE_DECODER_REGISTRY,
    class_decoder_registry: TClassDecoderRegistry = CORE_CLASS_DECODER_REGISTRY,
) -> GenerationStrategy:
    """Load generation strategy from JSON."""
    nodes = (
        object_from_json(
            generation_strategy_json.pop("nodes"),
            decoder_registry=decoder_registry,
            class_decoder_registry=class_decoder_registry,
        )
        if "nodes" in generation_strategy_json
        else []
    )

    steps = object_from_json(
        generation_strategy_json.pop("steps"),
        decoder_registry=decoder_registry,
        class_decoder_registry=class_decoder_registry,
    )
    if len(steps) > 0:
        gs = GenerationStrategy(steps=steps, name=generation_strategy_json.pop("name"))
        gs._curr = gs._steps[generation_strategy_json.pop("curr_index")]
    else:
        gs = GenerationStrategy(nodes=nodes, name=generation_strategy_json.pop("name"))
        curr_node_name = generation_strategy_json.pop("curr_node_name")
        for node in gs._nodes:
            if node.node_name == curr_node_name:
                gs._curr = node
                break

    gs._db_id = object_from_json(
        generation_strategy_json.pop("db_id"),
        decoder_registry=decoder_registry,
        class_decoder_registry=class_decoder_registry,
    )
    gs._experiment = experiment or object_from_json(
        generation_strategy_json.pop("experiment"),
        decoder_registry=decoder_registry,
        class_decoder_registry=class_decoder_registry,
    )
    gs._generator_runs = object_from_json(
        generation_strategy_json.pop("generator_runs"),
        decoder_registry=decoder_registry,
        class_decoder_registry=class_decoder_registry,
    )
    return gs


def surrogate_from_list_surrogate_json(
    list_surrogate_json: dict[str, Any],
    decoder_registry: TDecoderRegistry = CORE_DECODER_REGISTRY,
    class_decoder_registry: TClassDecoderRegistry = CORE_CLASS_DECODER_REGISTRY,
) -> Surrogate:
    logger.warning(
        "`ListSurrogate` has been deprecated. Reconstructing a `Surrogate` "
        "with as similar properties as possible."
    )
    if "submodel_input_transforms" in list_surrogate_json:
        (
            list_surrogate_json["submodel_input_transform_classes"],
            list_surrogate_json["submodel_input_transform_options"],
        ) = get_input_transform_json_components(
            list_surrogate_json.pop("submodel_input_transforms"),
            decoder_registry=decoder_registry,
            class_decoder_registry=class_decoder_registry,
        )
    if "submodel_outcome_transforms" in list_surrogate_json:
        (
            list_surrogate_json["submodel_outcome_transform_classes"],
            list_surrogate_json["submodel_outcome_transform_options"],
        ) = get_outcome_transform_json_components(
            list_surrogate_json.pop("submodel_outcome_transforms"),
            decoder_registry=decoder_registry,
            class_decoder_registry=class_decoder_registry,
        )
    return Surrogate(
        surrogate_spec=SurrogateSpec(
            model_configs=[
                ModelConfig(
                    botorch_model_class=object_from_json(
                        object_json=list_surrogate_json.get("botorch_submodel_class"),
                        decoder_registry=decoder_registry,
                        class_decoder_registry=class_decoder_registry,
                    ),
                    model_options=list_surrogate_json.get("submodel_options"),
                    mll_class=object_from_json(
                        object_json=list_surrogate_json.get("mll_class"),
                        decoder_registry=decoder_registry,
                        class_decoder_registry=class_decoder_registry,
                    ),
                    mll_options=list_surrogate_json.get("mll_options"),
                    input_transform_classes=object_from_json(
                        object_json=list_surrogate_json.get(
                            "submodel_input_transform_classes"
                        ),
                        decoder_registry=decoder_registry,
                        class_decoder_registry=class_decoder_registry,
                    ),
                    input_transform_options=object_from_json(
                        object_json=list_surrogate_json.get(
                            "submodel_input_transform_options"
                        ),
                        decoder_registry=decoder_registry,
                        class_decoder_registry=class_decoder_registry,
                    ),
                    outcome_transform_classes=object_from_json(
                        object_json=list_surrogate_json.get(
                            "submodel_outcome_transform_classes"
                        ),
                        decoder_registry=decoder_registry,
                        class_decoder_registry=class_decoder_registry,
                    ),
                    outcome_transform_options=object_from_json(
                        object_json=list_surrogate_json.get(
                            "submodel_outcome_transform_options"
                        ),
                        decoder_registry=decoder_registry,
                        class_decoder_registry=class_decoder_registry,
                    ),
                    covar_module_class=object_from_json(
                        object_json=list_surrogate_json.get(
                            "submodel_covar_module_class"
                        ),
                        decoder_registry=decoder_registry,
                        class_decoder_registry=class_decoder_registry,
                    ),
                    covar_module_options=list_surrogate_json.get(
                        "submodel_covar_module_options"
                    ),
                    likelihood_class=object_from_json(
                        object_json=list_surrogate_json.get(
                            "submodel_likelihood_class"
                        ),
                        decoder_registry=decoder_registry,
                        class_decoder_registry=class_decoder_registry,
                    ),
                    likelihood_options=list_surrogate_json.get(
                        "submodel_likelihood_options"
                    ),
                    name="from deprecated args",
                )
            ]
        )
    )


def get_input_transform_json_components(
    input_transforms_json: list[dict[str, Any]] | dict[str, Any] | None,
    decoder_registry: TDecoderRegistry = CORE_DECODER_REGISTRY,
    class_decoder_registry: TClassDecoderRegistry = CORE_CLASS_DECODER_REGISTRY,
) -> tuple[list[dict[str, Any]] | None, dict[str, Any] | None]:
    if input_transforms_json is None:
        return None, None
    if isinstance(input_transforms_json, dict):
        # This is a single input transform.
        input_transforms_json = [input_transforms_json]
    else:
        input_transforms_json = [
            input_transform_json
            for input_transform_json in input_transforms_json
            if input_transform_json is not None
        ]
    input_transform_classes_json = [
        input_transform_json["index"] for input_transform_json in input_transforms_json
    ]
    input_transform_options_json = {
        assert_is_instance(input_transform_json["__type"], str): input_transform_json[
            "state_dict"
        ]
        for input_transform_json in input_transforms_json
    }
    return input_transform_classes_json, input_transform_options_json


def get_outcome_transform_json_components(
    outcome_transforms_json: list[dict[str, Any]] | None,
    decoder_registry: TDecoderRegistry = CORE_DECODER_REGISTRY,
    class_decoder_registry: TClassDecoderRegistry = CORE_CLASS_DECODER_REGISTRY,
) -> tuple[list[dict[str, Any]] | None, dict[str, Any] | None]:
    if outcome_transforms_json is None:
        return None, None

    outcome_transforms_json = [
        outcome_transform_json
        for outcome_transform_json in outcome_transforms_json
        if outcome_transform_json is not None
    ]
    outcome_transform_classes_json = [
        outcome_transform_json["index"]
        for outcome_transform_json in outcome_transforms_json
    ]
    outcome_transform_options_json = {
        assert_is_instance(
            outcome_transform_json["__type"], str
        ): outcome_transform_json["state_dict"]
        for outcome_transform_json in outcome_transforms_json
    }
    return outcome_transform_classes_json, outcome_transform_options_json


def objective_from_json(
    object_json: dict[str, Any],
    decoder_registry: TDecoderRegistry = CORE_DECODER_REGISTRY,
    class_decoder_registry: TClassDecoderRegistry = CORE_CLASS_DECODER_REGISTRY,
) -> Objective:
    """Load an ``Objective`` from JSON in a backwards compatible way.

    If both ``minimize`` and ``lower_is_better`` are specified but have conflicting
    values, this will overwrite ``lower_is_better=minimize`` to resolve the conflict.

    # TODO: Do we need to do this for scalarized objective as well?
    """
    input_args = {
        k: object_from_json(
            v,
            decoder_registry=decoder_registry,
            class_decoder_registry=class_decoder_registry,
        )
        for k, v in object_json.items()
    }
    metric = input_args.pop("metric")
    minimize = input_args.pop("minimize")
    if metric.lower_is_better is not None and metric.lower_is_better != minimize:
        logger.warning(
            f"Metric {metric.name} has {metric.lower_is_better=} but objective "
            f"specifies {minimize=}. Overwriting ``lower_is_better`` to match "
            f"the optimization direction {minimize=}."
        )
        metric.lower_is_better = minimize
    return Objective(
        metric=metric,
        minimize=minimize,
        **input_args,  # For future compatibility.
    )


def _update_deprecated_model_registry(name: str) -> str:
    """Update the enum name for deprecated model registry entries to point to
    a replacement model. This will log an exception to alert the user to the change.

    The replacement models are listed in `_DEPRECATED_MODEL_TO_REPLACEMENT` above.
    If a deprecated model does not list a replacement, nothing will be done and it
    will error out while looking it up in the corresponding enum.

    Args:
        name: The name of the ``Generators`` enum.

    Returns:
        Either the given name or the name of a replacement ``Generators`` enum.
    """
    if name in _DEPRECATED_MODEL_TO_REPLACEMENT:
        new_name = _DEPRECATED_MODEL_TO_REPLACEMENT[name]
        logger.exception(
            f"{name} model is deprecated and replaced by Generators.{new_name}. "
            f"Please use {new_name} in the future. Note that this warning only "
            "enables deserialization of experiments with deprecated models. "
            "Model fitting with the loaded experiment may still fail. "
        )
        return new_name
    else:
        return name



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from unittest import mock

import numpy as np
from ax.adapter.random import RandomAdapter
from ax.adapter.registry import Cont_X_trans
from ax.core.arm import Arm
from ax.core.experiment import Experiment
from ax.core.metric import Metric
from ax.core.observation import ObservationFeatures
from ax.core.parameter import ParameterType, RangeParameter
from ax.core.parameter_constraint import (
    OrderConstraint,
    ParameterConstraint,
    SumConstraint,
)
from ax.core.search_space import SearchSpace
from ax.exceptions.core import SearchSpaceExhausted
from ax.generators.random.base import RandomGenerator
from ax.generators.random.sobol import SobolGenerator
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_data,
    get_search_space_for_range_values,
    get_small_discrete_search_space,
)


class RandomAdapterTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        x = RangeParameter("x", ParameterType.FLOAT, lower=0, upper=1)
        y = RangeParameter("y", ParameterType.FLOAT, lower=1, upper=2)
        z = RangeParameter("z", ParameterType.FLOAT, lower=0, upper=5)
        self.parameters = [x, y, z]
        parameter_constraints: list[ParameterConstraint] = [
            OrderConstraint(x, y),
            SumConstraint([x, z], False, 3.5),
        ]
        self.search_space = SearchSpace(self.parameters, parameter_constraints)
        self.experiment = Experiment(search_space=self.search_space)
        self.model_gen_options = {"option": "yes"}

    def test_fit(self) -> None:
        adapter = RandomAdapter(experiment=self.experiment, generator=RandomGenerator())
        self.assertEqual(adapter.parameters, ["x", "y", "z"])
        self.assertTrue(isinstance(adapter.generator, RandomGenerator))

    def test_predict(self) -> None:
        adapter = RandomAdapter(experiment=self.experiment, generator=RandomGenerator())
        with self.assertRaises(NotImplementedError):
            adapter._predict([])

    def test_cross_validate(self) -> None:
        adapter = RandomAdapter(experiment=self.experiment, generator=RandomGenerator())
        with self.assertRaises(NotImplementedError):
            # pyre-ignore[6]: None input for testing.
            adapter._cross_validate(self.search_space, None, None)

    def test_gen_w_constraints(self) -> None:
        adapter = RandomAdapter(experiment=self.experiment, generator=RandomGenerator())
        with mock.patch.object(
            adapter.generator,
            "gen",
            return_value=(
                np.array([[1.0, 2.0, 3.0], [3.0, 4.0, 3.0]]),
                np.array([1.0, 2.0]),
            ),
        ) as mock_gen:
            gen_results = adapter._gen(
                n=3,
                search_space=self.search_space,
                pending_observations={},
                fixed_features=ObservationFeatures({"z": 3.0}),
                optimization_config=None,
                # pyre-fixme[6]: For 6th param expected `Optional[Dict[str,
                # Union[None, Dict[str, typing.Any], OptimizationConfig,
                # AcquisitionFunction, float, int, str]]]` but got `Dict[str,
                # str]`.
                model_gen_options=self.model_gen_options,
            )
        gen_args = mock_gen.mock_calls[0][2]
        self.assertEqual(gen_args["n"], 3)
        self.assertEqual(gen_args["bounds"], [(0.0, 1.0), (1.0, 2.0), (0.0, 5.0)])
        self.assertTrue(
            np.array_equal(
                gen_args["linear_constraints"][0],
                np.array([[1.0, -1, 0.0], [-1.0, 0.0, -1.0]]),
            )
        )
        self.assertTrue(
            np.array_equal(gen_args["linear_constraints"][1], np.array([[0.0], [-3.5]]))
        )
        self.assertEqual(gen_args["fixed_features"], {2: 3.0})
        self.assertEqual(gen_args["model_gen_options"], {"option": "yes"})
        obsf = gen_results.observation_features
        self.assertEqual(obsf[0].parameters, {"x": 1.0, "y": 2.0, "z": 3.0})
        self.assertEqual(obsf[1].parameters, {"x": 3.0, "y": 4.0, "z": 3.0})
        self.assertTrue(np.array_equal(gen_results.weights, np.array([1.0, 2.0])))

    def test_gen_simple(self) -> None:
        # Test with no constraints, no fixed feature, no pending observations
        search_space = SearchSpace(self.parameters[:2])
        adapter = RandomAdapter(
            experiment=Experiment(search_space=search_space),
            generator=RandomGenerator(),
        )
        with mock.patch.object(
            adapter.generator,
            "gen",
            return_value=(np.array([[1.0, 2.0], [3.0, 4.0]]), np.array([1.0, 2.0])),
        ) as mock_gen:
            adapter._gen(
                n=3,
                search_space=search_space,
                pending_observations={},
                fixed_features=ObservationFeatures({}),
                optimization_config=None,
                # pyre-fixme[6]: For 6th param expected `Optional[Dict[str,
                # Union[None, Dict[str, typing.Any], OptimizationConfig,
                # AcquisitionFunction, float, int, str]]]` but got `Dict[str,
                # str]`.
                model_gen_options=self.model_gen_options,
            )
        gen_args = mock_gen.mock_calls[0][2]
        self.assertEqual(gen_args["bounds"], [(0.0, 1.0), (1.0, 2.0)])
        self.assertIsNone(gen_args["linear_constraints"])
        self.assertIsNone(gen_args["fixed_features"])

    def test_deduplicate(self) -> None:
        exp = Experiment(search_space=get_small_discrete_search_space())
        sobol = RandomAdapter(
            experiment=exp,
            generator=SobolGenerator(deduplicate=True),
            transforms=Cont_X_trans,
        )
        for _ in range(4):  # Search space is {[0, 1], {"red", "panda"}}
            # Generate & attach trials to the experiment so that the
            # generated points are used for deduplication.
            gr = sobol.gen(1)
            exp.new_trial(generator_run=gr).mark_running(no_runner_required=True)
            self.assertEqual(len(gr.arms), 1)
        with self.assertRaises(SearchSpaceExhausted):
            sobol.gen(1)

    def test_search_space_not_expanded(self) -> None:
        data = get_data(num_non_sq_arms=0)
        sq_arm = Arm(name="status_quo", parameters={"x": 10.0, "y": 1.0, "z": 1.0})
        experiment = Experiment(
            search_space=self.search_space,
            status_quo=sq_arm,
        )
        trial = experiment.new_trial()
        trial.add_arm(sq_arm)
        trial.mark_running(no_runner_required=True)
        trial.mark_completed()
        experiment.add_tracking_metric(metric=Metric("ax_test_metric"))
        sobol = RandomAdapter(
            search_space=self.search_space,
            generator=SobolGenerator(),
            experiment=experiment,
            data=data,
            transforms=Cont_X_trans,
        )
        # test that search space is not expanded
        sobol.gen(1)
        self.assertEqual(sobol._model_space, sobol._search_space)

    def test_generated_points(self) -> None:
        # Checks for generated points argument passed to Generator.gen.
        # Search space has two range parameters in [0, 5].
        exp = Experiment(
            search_space=get_search_space_for_range_values(min=0.0, max=5.0)
        )
        generator = SobolGenerator(deduplicate=True)
        gen_res = generator.gen(
            n=1, bounds=[(0.0, 1.0), (0.0, 1.0)], rounding_func=lambda x: x
        )
        # Using Cont_X_trans, particularly UnitX here to test transform application.
        adapter = RandomAdapter(
            experiment=exp, generator=generator, transforms=Cont_X_trans
        )

        # No pending points or previous trials on the experiment.
        with mock.patch.object(generator, "gen", return_value=gen_res) as mock_gen:
            adapter.gen(n=1)
        self.assertIsNone(mock_gen.call_args.kwargs["generated_points"])

        # Attach two trials to the experiment.
        exp.new_trial().add_arm(Arm(parameters={"x": 0.0, "y": 0.0})).mark_running(
            no_runner_required=True
        )
        exp.new_trial().add_arm(Arm(parameters={"x": 2.0, "y": 2.0})).mark_running(
            no_runner_required=True
        )
        with mock.patch.object(generator, "gen", return_value=gen_res) as mock_gen:
            adapter.gen(n=1)
        self.assertEqual(
            mock_gen.call_args.kwargs["generated_points"].tolist(),
            [[0.0, 0.0], [0.4, 0.4]],
        )

        # Add pending points -- only unique ones should be passed down.
        pending_observations = {
            m: [ObservationFeatures(parameters={"x": 3.0, "y": 3.0})]
            for m in ("m1", "m2")
        }
        with mock.patch.object(generator, "gen", return_value=gen_res) as mock_gen:
            adapter.gen(n=1, pending_observations=pending_observations)
        self.assertEqual(
            mock_gen.call_args.kwargs["generated_points"].tolist(),
            [[0.0, 0.0], [0.4, 0.4], [0.6, 0.6]],
        )

        # Turn off deduplicate, nothing should be passed down.
        generator.deduplicate = False
        with mock.patch.object(generator, "gen", return_value=gen_res) as mock_gen:
            adapter.gen(n=1, pending_observations=pending_observations)
        self.assertIsNone(mock_gen.call_args.kwargs["generated_points"])

        # Test filtering out-of-design arms during deduplication
        # Create experiment with in-design and out-of-design arms
        exp_with_ood_arms = Experiment(
            search_space=get_search_space_for_range_values(min=0.0, max=5.0)
        )
        in_design_arm = Arm(
            name="in_design", parameters={"x": 2.0, "y": 3.0}
        )  # Within [0, 5]
        out_of_design_arm = Arm(
            name="out_of_design", parameters={"x": 6.0, "y": 7.0}
        )  # Outside [0, 5]

        exp_with_ood_arms.new_trial().add_arm(in_design_arm).mark_running(
            no_runner_required=True
        )
        exp_with_ood_arms.new_trial().add_arm(out_of_design_arm).mark_running(
            no_runner_required=True
        )

        generator = SobolGenerator(deduplicate=True)
        adapter_mixed = RandomAdapter(
            experiment=exp_with_ood_arms,
            generator=generator,
            transforms=Cont_X_trans,
        )

        # Only the in-design arm should be included in generated_points
        with mock.patch.object(generator, "gen", return_value=gen_res) as mock_gen:
            adapter_mixed.gen(n=1)

        generated_points = mock_gen.call_args.kwargs["generated_points"]
        self.assertEqual(len(generated_points), 1)

        # Test case where all arms are out-of-design
        exp_all_out_of_design = Experiment(
            search_space=get_search_space_for_range_values(min=0.0, max=5.0)
        )
        out_of_design_arm1 = Arm(name="out1", parameters={"x": 6.0, "y": 7.0})
        out_of_design_arm2 = Arm(name="out2", parameters={"x": -1.0, "y": 8.0})

        exp_all_out_of_design.new_trial().add_arm(out_of_design_arm1).mark_running(
            no_runner_required=True
        )
        exp_all_out_of_design.new_trial().add_arm(out_of_design_arm2).mark_running(
            no_runner_required=True
        )

        generator_all_out = SobolGenerator(deduplicate=True)
        adapter_all_out = RandomAdapter(
            experiment=exp_all_out_of_design,
            generator=generator_all_out,
            transforms=Cont_X_trans,
        )

        # When all arms are out-of-design, generated_points should be empty
        with mock.patch.object(
            generator_all_out, "gen", return_value=gen_res
        ) as mock_gen:
            adapter_all_out.gen(n=1)

        generated_points_all_out = mock_gen.call_args.kwargs["generated_points"]
        self.assertIsNone(generated_points_all_out)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
Mixed integer extensions of some common synthetic test functions.
These are adapted from [Daulton2022bopr]_.

References

.. [Daulton2022bopr]
    S. Daulton, X. Wan, D. Eriksson, M. Balandat, M. A. Osborne, E. Bakshy.
    Bayesian Optimization over Discrete and Mixed Spaces via Probabilistic
    Reparameterization. Advances in Neural Information Processing Systems
    35, 2022.
"""

from ax.benchmark.benchmark_problem import BenchmarkProblem, get_soo_opt_config
from ax.benchmark.benchmark_test_functions.botorch_test import BoTorchTestFunction
from ax.core.parameter import ParameterType, RangeParameter
from ax.core.search_space import SearchSpace
from botorch.test_functions.synthetic import Ackley, Hartmann, Rosenbrock


def _get_problem_from_common_inputs(
    *,
    bounds: list[tuple[float, float]],
    dim_int: int,
    metric_name: str,
    lower_is_better: bool,
    observe_noise_sd: bool,
    test_problem_class: type[Hartmann | Ackley | Rosenbrock],
    benchmark_name: str,
    num_trials: int,
    optimal_value: float,
    baseline_value: float,
    test_problem_bounds: list[tuple[float, float]] | None = None,
) -> BenchmarkProblem:
    """This is a helper that deduplicates common bits of the below problems.

    Args:
        bounds: The parameter bounds. These will be passed to
            `BotorchTestFunction` as `modified_bounds`, and the parameters
            will be renormalized from these bounds to the bounds of the original
            problem. For example, if `bounds` are [(0, 3)] and the test
            problem's original bounds are [(0, 2)], then the original problem
            can be evaluated at 0, 2/3, 4/3, and 2.
        dim_int: The number of integer dimensions. First `dim_int` parameters
            are assumed to be integers.
        metric_name: The name of the metric.
        lower_is_better: If true, the goal is to minimize the metric.
        observe_noise_sd: Whether to report the standard deviation of the
            observation noise.
        test_problem_class: The BoTorch test problem class.
        benchmark_name: The name of the benchmark problem.
        num_trials: The number of trials.
        optimal_value: Best attainable value, if known. If unknown, as may be
            the case for mixed-integer problems, choose a value that is known to
            be at least as good as the true optimum, to prevent benchmarks from
            attaining scores of over 100%. One strategy for choosing this value
            is to choose the overall optimum of the problem without regard to
            the integer restrictions.
        baseline_value: Will be passed to the `BenchmarkProblem`.
        test_problem_bounds: Optional bounds to evaluate the base test problem on.
            These are passed in as `bounds` while initializing the test problem.

    Returns:
        A mixed-integer BenchmarkProblem constructed from the given inputs.
    """
    dim = len(bounds)
    search_space = SearchSpace(
        parameters=[
            RangeParameter(
                name=f"x{i + 1}",
                parameter_type=(
                    ParameterType.INT if i < dim_int else ParameterType.FLOAT
                ),
                lower=bounds[i][0],
                upper=bounds[i][1],
            )
            for i in range(dim)
        ]
    )
    optimization_config = get_soo_opt_config(
        outcome_names=[metric_name],
        lower_is_better=lower_is_better,
        observe_noise_sd=observe_noise_sd,
    )

    if test_problem_bounds is None:
        test_problem = test_problem_class(dim=dim)
    else:
        test_problem = test_problem_class(dim=dim, bounds=test_problem_bounds)
    test_function = BoTorchTestFunction(
        botorch_problem=test_problem,
        modified_bounds=bounds,
        outcome_names=[metric_name],
    )
    return BenchmarkProblem(
        name=benchmark_name + ("_observed_noise" if observe_noise_sd else ""),
        search_space=search_space,
        optimization_config=optimization_config,
        test_function=test_function,
        num_trials=num_trials,
        optimal_value=optimal_value,
        baseline_value=baseline_value,
    )


def get_discrete_hartmann(
    num_trials: int = 50,
    observe_noise_sd: bool = False,
    bounds: list[tuple[float, float]] | None = None,
) -> BenchmarkProblem:
    """6D Hartmann problem where first 4 dimensions are discretized."""
    dim_int = 4
    if bounds is None:
        bounds = [
            (0, 3),
            (0, 3),
            (0, 19),
            (0, 19),
            (0.0, 1.0),
            (0.0, 1.0),
        ]
    return _get_problem_from_common_inputs(
        bounds=bounds,
        dim_int=dim_int,
        metric_name="Hartmann",
        lower_is_better=True,
        observe_noise_sd=observe_noise_sd,
        test_problem_class=Hartmann,
        benchmark_name="Discrete Hartmann",
        num_trials=num_trials,
        # The best value we've found so far on this mixed problem is -2.89. The
        # optimum without regards to the integer constraints is -3.3224, but
        # that won't be attainable here.
        optimal_value=-3.0,
        # Baseline values were obtained with `compute_baseline_value_from_sobol`
        baseline_value=-0.6755184773211834,
    )


def get_discrete_ackley(
    num_trials: int = 50,
    observe_noise_sd: bool = False,
    bounds: list[tuple[float, float]] | None = None,
) -> BenchmarkProblem:
    """13D Ackley problem where first 10 dimensions are discretized.

    This also restricts Ackley evaluation bounds to [0, 1].
    """
    dim = 13
    dim_int = 10
    if bounds is None:
        bounds = [
            *[(0, 2)] * 5,
            *[(0, 4)] * 5,
            *[(0.0, 1.0)] * 3,
        ]
    return _get_problem_from_common_inputs(
        bounds=bounds,
        dim_int=dim_int,
        metric_name="Ackley",
        lower_is_better=True,
        observe_noise_sd=observe_noise_sd,
        test_problem_class=Ackley,
        benchmark_name="Discrete Ackley",
        num_trials=num_trials,
        test_problem_bounds=[(0.0, 1.0)] * dim,
        # Ackley's lowest value is at (0, 0, ..., 0), which is in the search
        # space, so the restriction to integers doesn't change the optimum
        optimal_value=0.0,
        baseline_value=3.1869268815137968,
    )


def get_discrete_rosenbrock(
    num_trials: int = 50,
    observe_noise_sd: bool = False,
    bounds: list[tuple[float, float]] | None = None,
) -> BenchmarkProblem:
    """10D Rosenbrock problem where first 6 dimensions are discretized."""
    dim_int = 6
    if bounds is None:
        bounds = [
            *[(0, 3)] * 6,
            *[(0.0, 1.0)] * 4,
        ]
    return _get_problem_from_common_inputs(
        bounds=bounds,
        dim_int=dim_int,
        metric_name="Rosenbrock",
        lower_is_better=True,
        observe_noise_sd=observe_noise_sd,
        test_problem_class=Rosenbrock,
        benchmark_name="Discrete Rosenbrock",
        num_trials=num_trials,
        # Rosenbrock's lowest value is at (1, 1, ..., 1), which is in the search
        # space, so the restriction to integers doesn't change the optimum
        optimal_value=0.0,
        baseline_value=705714.2851460224,
    )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import logging
from abc import ABC, abstractmethod
from collections.abc import Iterable, Sequence
from logging import Logger

import pandas as pd
from ax.core.batch_trial import BatchTrial
from ax.core.experiment import Experiment
from ax.core.map_data import MAP_KEY, MapData
from ax.core.objective import MultiObjective
from ax.core.trial_status import TrialStatus
from ax.early_stopping.utils import estimate_early_stopping_savings
from ax.generation_strategy.generation_node import GenerationNode
from ax.utils.common.base import Base
from ax.utils.common.logger import get_logger
from pyre_extensions import assert_is_instance, none_throws

logger: Logger = get_logger(__name__)


class BaseEarlyStoppingStrategy(ABC, Base):
    """Interface for heuristics that halt trials early, typically based on early
    results from that trial."""

    def __init__(
        self,
        metric_names: Iterable[str] | None = None,
        min_progression: float | None = None,
        max_progression: float | None = None,
        min_curves: int | None = None,
        trial_indices_to_ignore: list[int] | None = None,
        normalize_progressions: bool = False,
    ) -> None:
        """A BaseEarlyStoppingStrategy class.

        Args:
            metric_names: The names of the metrics the strategy will interact with.
                If no metric names are provided, considers the objective metric(s).
            min_progression: Only stop trials if the latest progression value
                (e.g. timestamp, epochs, training data used) is greater than this
                threshold. Prevents stopping prematurely before enough data is gathered
                to make a decision.
            max_progression: Do not stop trials that have passed `max_progression`.
                Useful if we prefer finishing a trial that are already near completion.
            min_curves: Trials will not be stopped until a number of trials
                `min_curves` have completed with curve data attached. That is, if
                `min_curves` trials are completed but their curve data was not
                successfully retrieved, further trials may not be early-stopped.
            trial_indices_to_ignore: Trial indices that should not be early stopped.
            normalize_progressions: Normalizes the progression column of the MapData df
                by dividing by the max. If the values were originally in [0, `prog_max`]
                (as we would expect), the transformed values will be in [0, 1]. Useful
                for inferring the max progression and allows `min_progression` to be
                specified in the transformed space. IMPORTANT: Typically, `min_curves`
                should be > 0 to ensure that at least one trial has completed and that
                we have a reliable approximation for `prog_max`.
        """
        self.metric_names = metric_names
        self.min_progression = min_progression
        self.max_progression = max_progression
        self.min_curves = min_curves
        self.trial_indices_to_ignore = trial_indices_to_ignore
        self.normalize_progressions = normalize_progressions

    @abstractmethod
    def should_stop_trials_early(
        self,
        trial_indices: set[int],
        experiment: Experiment,
        current_node: GenerationNode | None = None,
    ) -> dict[int, str | None]:
        """Decide whether to complete trials before evaluation is fully concluded.

        Typical examples include stopping a machine learning model's training, or
        halting the gathering of samples before some planned number are collected.

        Args:
            trial_indices: Indices of candidate trials to stop early.
            experiment: Experiment that contains the trials and other contextual data.
            current_node: The current ``GenerationNode`` on the ``GenerationStrategy``
                used to generate trials for the ``Experiment``. Early stopping
                strategies may utilize components of the current node when making
                stopping decisions.

        Returns:
            A dictionary mapping trial indices that should be early stopped to
            (optional) messages with the associated reason.
        """
        pass  # pragma: nocover

    def estimate_early_stopping_savings(self, experiment: Experiment) -> float:
        """Estimate early stopping savings using progressions of the MapMetric present
        on the EarlyStoppingConfig as a proxy for resource usage.

        Args:
            experiment: The experiment containing the trials and metrics used
                to estimate early stopping savings.

        Returns:
            The estimated resource savings as a fraction of total resource usage (i.e.
            0.11 estimated savings indicates we would expect the experiment to have used
            11% more resources without early stopping present)
        """
        if experiment.default_data_constructor is not MapData:
            return 0

        return estimate_early_stopping_savings(experiment=experiment)

    def _check_validity_and_get_data(
        self, experiment: Experiment, metric_names: list[str]
    ) -> MapData | None:
        """Validity checks and returns the `MapData` used for early stopping that
        is associated with `metric_names`. This function also handles normalizing
        progressions.
        """
        data = experiment.lookup_data()
        if data.df.empty:
            logger.info(
                f"{self.__class__.__name__} received empty data. "
                "Not stopping any trials."
            )
            return None
        for metric_name in metric_names:
            if metric_name not in set(data.df["metric_name"]):
                logger.info(
                    f"{self.__class__.__name__} did not receive data from the "
                    f"objective metric `{metric_name}`. Not stopping any trials."
                )
                return None

        if not isinstance(data, MapData):
            logger.info(
                f"{self.__class__.__name__} expects MapData, but the "
                f"data attached to experiment is of type {type(data)}. "
                "Not stopping any trials."
            )
            return None

        data = assert_is_instance(data, MapData)
        map_df = data.map_df
        # keep only relevant metrics
        map_df = map_df[map_df["metric_name"].isin(metric_names)].copy()
        if self.normalize_progressions:
            values = map_df[MAP_KEY].astype(float)
            map_df[MAP_KEY] = values / values.abs().max()
        return MapData(df=map_df)

    @staticmethod
    def _log_and_return_trial_ignored(
        logger: logging.Logger, trial_index: int
    ) -> tuple[bool, str]:
        """Helper function for logging/constructing a reason when a trial
        should be ignored."""
        logger.info(
            f"Trial {trial_index} should be ignored and not considered "
            "for early stopping."
        )
        return False, "Specified as a trial to be ignored for early stopping."

    @staticmethod
    def _log_and_return_no_data(
        logger: logging.Logger, trial_index: int, metric_name: str
    ) -> tuple[bool, str]:
        """Helper function for logging/constructing a reason when there is no data."""
        logger.info(
            f"There is not yet any data associated with trial {trial_index} and "
            f"metric {metric_name}. Not early stopping this trial."
        )
        return False, "No data available to make an early stopping decision."

    @staticmethod
    def _log_and_return_progression_range(
        logger: logging.Logger,
        trial_index: int,
        trial_last_progression: float,
        min_progression: float | None,
        max_progression: float | None,
        metric_name: str,
    ) -> tuple[bool, str]:
        """Helper function for logging/constructing a reason when min progression
        is not yet reached."""
        reason = (
            f"Most recent progression ({trial_last_progression}) that is available for "
            f"metric {metric_name} falls out of the "
            f"min/max_progression range ({min_progression}, {max_progression})."
        )
        logger.info(
            f"Trial {trial_index}'s m{reason[1:]} Not early stopping this trial."
        )
        return False, reason

    def is_eligible_any(
        self,
        trial_indices: set[int],
        experiment: Experiment,
        df: pd.DataFrame,
    ) -> bool:
        """Perform a series of default checks for a set of trials `trial_indices` and
        determine if at least one of them is eligible for further stopping logic:
            1. Check that at least `self.min_curves` trials have completed`
            2. Check that at least one trial has reached `self.min_progression`
        Returns a boolean indicating if all checks are passed.

        This is useful for some situations where if no trials are eligible for stopping,
        then we can skip costly steps, such as model fitting, that occur before
        individual trials are considered for stopping.
        """
        # check for batch trials
        for idx, trial in experiment.trials.items():
            if isinstance(trial, BatchTrial):
                # In particular, align_partial_results requires a 1-1 mapping between
                # trial indices and arm names, which is not the case for batch trials.
                # See align_partial_results for more details.
                raise ValueError(
                    f"Trial {idx} is a BatchTrial, which is not yet supported by "
                    "early stopping strategies."
                )

        # check that there are sufficient completed trials
        num_completed = len(experiment.trial_indices_by_status[TrialStatus.COMPLETED])
        if self.min_curves is not None and num_completed < self.min_curves:
            return False

        # check that at least one trial has reached `self.min_progression`
        df_trials = df[df["trial_index"].isin(trial_indices)].dropna(subset=["mean"])
        any_last_prog = 0
        if not df_trials[MAP_KEY].empty:
            any_last_prog = df_trials[MAP_KEY].max()

        if self.min_progression is not None and any_last_prog < self.min_progression:
            # No trials have reached `self.min_progression`, not stopping any trials
            return False

        return True

    def is_eligible(
        self,
        trial_index: int,
        experiment: Experiment,
        df: pd.DataFrame,
    ) -> tuple[bool, str | None]:
        """Perform a series of default checks for a specific trial `trial_index` and
        determines whether it is eligible for further stopping logic:
            1. Check for ignored indices based on `self.trial_indices_to_ignore`
            2. Check that `df` contains data for the trial `trial_index`
            3. Check that the trial has reached `self.min_progression`
            4. Check that the trial hasn't surpassed `self.max_progression`
        Returns two elements: a boolean indicating if all checks are passed and a
        str indicating the reason that early stopping is not applied (None if all
        checks pass).

        Args:
            trial_index: The index of the trial to check.
            experiment: The experiment containing the trial.
            df: A dataframe containing the time-dependent metrics for the trial.
                NOTE: `df` should only contain data with `metric_name` fields that are
                associated with the early stopping strategy. This is usually done
                automatically in `_check_validity_and_get_data`. `is_eligible` might
                otherwise return False even though the trial is eligible, if there are
                secondary tracking metrics that are in `df` but shouldn't be considered
                in the early stopping decision.

        Returns:
            A tuple of two elements: a boolean indicating if the trial is eligible and
                an optional string indicating any reason for ineligiblity.
        """
        # check for ignored indices
        if (
            self.trial_indices_to_ignore is not None
            and trial_index in self.trial_indices_to_ignore
        ):
            return self._log_and_return_trial_ignored(
                logger=logger, trial_index=trial_index
            )

        # Check eligibility of each metric.
        for metric_name, metric_df in df.groupby("metric_name"):
            # check for no data
            df_trial = metric_df[metric_df["trial_index"] == trial_index]
            df_trial = df_trial.dropna(subset=["mean"])
            if df_trial.empty:
                return self._log_and_return_no_data(
                    logger=logger, trial_index=trial_index, metric_name=metric_name
                )

            # check for min/max progression
            trial_last_prog = df_trial[MAP_KEY].max()
            if (
                self.min_progression is not None
                and trial_last_prog < self.min_progression
            ) or (
                self.max_progression is not None
                and trial_last_prog > self.max_progression
            ):
                return self._log_and_return_progression_range(
                    logger=logger,
                    trial_index=trial_index,
                    trial_last_progression=trial_last_prog,
                    min_progression=self.min_progression,
                    max_progression=self.max_progression,
                    metric_name=metric_name,
                )
        return True, None

    def _default_objective_and_direction(
        self, experiment: Experiment
    ) -> tuple[str, bool]:
        objectives_to_directions = self._all_objectives_and_directions(
            experiment=experiment
        )
        # if it is a multi-objective optimization problem, infer as first objective
        # although it is recommended to specify metric names explicitly.
        return next(iter(objectives_to_directions.items()))

    def _all_objectives_and_directions(self, experiment: Experiment) -> dict[str, bool]:
        """A dictionary mapping metric names to corresponding directions, i.e.
        a Boolean indicating whether the objective is minimized, for each objective in
        the experiment or in `self.metric_names`, if specified.

        Args:
            experiment: The experiment containing the optimization config.

        Returns: A dictionary mapping metric names to a Boolean indicating whether
            the objective is being minimized.
        """
        if self.metric_names is None:
            logger.debug(
                "No metric names specified. Defaulting to the objective metric(s).",
                stacklevel=2,
            )
            optimization_config = none_throws(experiment.optimization_config)
            objective = optimization_config.objective
            objectives = (
                objective.objectives
                if isinstance(objective, MultiObjective)
                else [objective]
            )
            directions = {}
            for objective in objectives:
                metric_name = objective.metric.name
                directions[metric_name] = objective.minimize

        else:
            metric_names = list(self.metric_names)
            directions = {}
            for metric_name in metric_names:
                minimize = experiment.metrics[metric_name].lower_is_better or False
                directions[metric_name] = minimize

        return directions


class ModelBasedEarlyStoppingStrategy(BaseEarlyStoppingStrategy):
    """A base class for model based early stopping strategies. Includes
    a helper function for processing MapData into arrays."""

    def __init__(
        self,
        metric_names: Iterable[str] | None = None,
        min_progression: float | None = None,
        max_progression: float | None = None,
        min_curves: int | None = None,
        trial_indices_to_ignore: list[int] | None = None,
        normalize_progressions: bool = False,
        min_progression_modeling: float | None = None,
    ) -> None:
        """A ModelBasedEarlyStoppingStrategy class.

        Args:
            metric_names: The names of the metrics the strategy will interact with.
                If no metric names are provided the objective metric is assumed.
            min_progression: Only stop trials if the latest progression value
                (e.g. timestamp, epochs, training data used) is greater than this
                threshold. Prevents stopping prematurely before enough data is gathered
                to make a decision.
            max_progression: Do not stop trials that have passed `max_progression`.
                Useful if we prefer finishing a trial that are already near completion.
            min_curves: Trials will not be stopped until a number of trials
                `min_curves` have completed with curve data attached. That is, if
                `min_curves` trials are completed but their curve data was not
                successfully retrieved, further trials may not be early-stopped.
            trial_indices_to_ignore: Trial indices that should not be early stopped.
            normalize_progressions: Normalizes the progression column of the MapData df
                by dividing by the max. If the values were originally in [0, `prog_max`]
                (as we would expect), the transformed values will be in [0, 1]. Useful
                for inferring the max progression and allows `min_progression` to be
                specified in the transformed space. IMPORTANT: Typically, `min_curves`
                should be > 0 to ensure that at least one trial has completed and that
                we have a reliable approximation for `prog_max`.
            min_progression_modeling: If set, this will exclude progressions that are
                below this threshold from being used or modeling. Useful when early data
                is not reliable or excessively noisy.
        """
        super().__init__(
            metric_names=metric_names,
            min_progression=min_progression,
            max_progression=max_progression,
            min_curves=min_curves,
            trial_indices_to_ignore=trial_indices_to_ignore,
            normalize_progressions=normalize_progressions,
        )
        self.min_progression_modeling = min_progression_modeling

    def _check_validity_and_get_data(
        self, experiment: Experiment, metric_names: list[str]
    ) -> MapData | None:
        """Validity checks and returns the `MapData` used for early stopping that
        is associated with `metric_names`. This function also handles normalizing
        progressions.
        """
        map_data = super()._check_validity_and_get_data(
            experiment=experiment, metric_names=metric_names
        )
        if map_data is not None and self.min_progression_modeling is not None:
            map_df = map_data.map_df
            map_df = map_df[map_df[MAP_KEY] >= self.min_progression_modeling]
            map_data = MapData(df=map_df)
        return map_data

    def get_training_data(
        self,
        experiment: Experiment,
        map_data: MapData,
        max_training_size: int | None = None,
        outcomes: Sequence[str] | None = None,
        parameters: list[str] | None = None,
    ) -> None:
        # Deprecated in Ax 1.1.0, so should be removed in Ax 1.2.0+.
        raise DeprecationWarning(
            "`ModelBasedEarlyStoppingStrategy.get_training_data` is deprecated. "
            "Subclasses should either extract the training data manually, "
            "or rely on the fitted surrogates available in the current generation "
            "node that is passed into `should_stop_trials_early`."
        )



#!/usr/bin/env fbpython
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

import dataclasses
import functools
import warnings
from typing import Any
from unittest import mock
from unittest.mock import Mock

import torch
from ax.core.search_space import SearchSpaceDigest
from ax.generators.torch.botorch_modular.acquisition import Acquisition
from ax.generators.torch.botorch_modular.sebo import (
    clamp_to_target,
    get_batch_initial_conditions,
    L1_norm_func,
    SEBOAcquisition,
)
from ax.generators.torch.botorch_modular.surrogate import Surrogate
from ax.generators.torch_base import TorchOptConfig
from ax.utils.common.constants import Keys
from ax.utils.common.testutils import TestCase
from ax.utils.testing.mock import mock_botorch_optimize
from botorch.acquisition.multi_objective.monte_carlo import (
    qNoisyExpectedHypervolumeImprovement,
)
from botorch.acquisition.penalized import L0Approximation
from botorch.models.deterministic import GenericDeterministicModel
from botorch.models.gp_regression import SingleTaskGP
from botorch.models.model import ModelList
from botorch.utils.datasets import SupervisedDataset
from pyre_extensions import none_throws


SEBOACQUISITION_PATH: str = SEBOAcquisition.__module__
ACQUISITION_PATH: str = Acquisition.__module__
CURRENT_PATH: str = __name__
SURROGATE_PATH: str = Surrogate.__module__


class TestSebo(TestCase):
    @mock_botorch_optimize
    def setUp(self) -> None:
        super().setUp()
        tkwargs: dict[str, Any] = {"dtype": torch.double}
        self.surrogate = Surrogate()
        # Function is f(x) = x_1 on [0, 1]^3, target point of [0.5, 0.5, 0.5].
        # Optimal soln is [1.0, 0.5, 0.5]
        self.X = torch.tensor(
            [
                [0.0, 0.0, 0.0],
                [0.6, 0.0, 0.0],
                [0.9, 0.0, 0.0],
                [0.0, 0.5, 0.0],
                [0.0, 1.0, 0.0],
                [0.0, 0.0, 0.5],
                [0.0, 0.0, 1.0],
                [0.9, 1.0, 1.0],
                [0.9, 0.0, 1.0],
                [0.9, 1.0, 0.0],
            ],
            **tkwargs,
        )
        self.target_point = torch.tensor([0.5, 0.5, 0.5], **tkwargs)
        self.Y = self.X[:, 0].unsqueeze(-1)
        self.Yvar = 0.1 * torch.ones_like(self.Y)
        self.training_data = [
            SupervisedDataset(
                X=self.X, Y=self.Y, feature_names=["a", "b", "c"], outcome_names=["m1"]
            )
        ]
        self.search_space_digest = SearchSpaceDigest(
            feature_names=["a", "b", "c"],
            bounds=[(0.0, 1.0), (0.0, 1.0), (0.0, 1.0)],
            target_values={2: 1.0},
        )
        self.surrogate.fit(
            datasets=self.training_data,
            search_space_digest=self.search_space_digest,
        )

        self.botorch_acqf_class = qNoisyExpectedHypervolumeImprovement
        self.objective_weights = torch.tensor([1.0], **tkwargs)
        # new transformed objective weights
        self.objective_weights_sebo = torch.tensor([1.0, -1.0], **tkwargs)
        self.objective_thresholds = torch.tensor([1.0], **tkwargs)
        self.objective_thresholds_sebo = torch.tensor([1.0, 3.0], **tkwargs)

        self.pending_observations = [torch.tensor([[0.5, 0.5, 0.5]], **tkwargs)]
        self.outcome_constraints = (
            torch.tensor([[1.0]], **tkwargs),
            torch.tensor([[2.0]], **tkwargs),
        )
        self.outcome_constraints_sebo = (
            torch.tensor([[1.0, 0.0]], **tkwargs),
            torch.tensor([[2.0]], **tkwargs),
        )
        self.linear_constraints = None
        self.fixed_features = {1: 1.0}
        self.options = {"target_point": self.target_point}
        self.inequality_constraints = [
            (torch.tensor([0, 1], **tkwargs), torch.tensor([1.0, -1.0], **tkwargs), 0)
        ]
        self.rounding_func = lambda x: x
        self.optimizer_options = {
            Keys.NUM_RESTARTS: 40,
            Keys.RAW_SAMPLES: 1024,
            "num_homotopy_steps": 3,
        }
        self.tkwargs = tkwargs
        self.torch_opt_config = TorchOptConfig(
            objective_weights=self.objective_weights,
            objective_thresholds=self.objective_thresholds,
            pending_observations=self.pending_observations,
            outcome_constraints=self.outcome_constraints,
            linear_constraints=self.linear_constraints,
            fixed_features=self.fixed_features,
        )
        self.torch_opt_config_2 = TorchOptConfig(
            objective_weights=self.objective_weights,
            objective_thresholds=self.objective_thresholds,
            pending_observations=self.pending_observations,
        )

    def get_acquisition_function(
        self,
        fixed_features: dict[int, float] | None = None,
        options: dict[str, str | float] | None = None,
        torch_opt_config: TorchOptConfig | None = None,
        botorch_acqf_options: dict[str, Any] | None = None,
    ) -> SEBOAcquisition:
        return SEBOAcquisition(
            botorch_acqf_class=qNoisyExpectedHypervolumeImprovement,
            surrogate=self.surrogate,
            search_space_digest=self.search_space_digest,
            torch_opt_config=dataclasses.replace(
                torch_opt_config or self.torch_opt_config,
                fixed_features=fixed_features or {},
            ),
            options=options or self.options,
            botorch_acqf_options=botorch_acqf_options or {},
        )

    def test_init(self) -> None:
        acquisition1 = self.get_acquisition_function(
            options={"target_point": self.target_point},
        )
        # Check that determinstic metric is added to surrogate
        surrogate = acquisition1.surrogate
        model_list = none_throws(surrogate._model)
        self.assertIsInstance(model_list, ModelList)
        # pyre-fixme[29]: `Union[(self: TensorBase, indices: Union[None, slice[Any, A...
        self.assertIsInstance(model_list.models[0], SingleTaskGP)
        # pyre-fixme[29]: `Union[(self: TensorBase, indices: Union[None, slice[Any, A...
        self.assertIsInstance(model_list.models[1], GenericDeterministicModel)

        # Check right penalty term is instantiated
        self.assertEqual(acquisition1.penalty_name, "L0_norm")
        # pyre-fixme[29]: `Union[(self: TensorBase, indices: Union[None, slice[Any, A...
        self.assertIsInstance(model_list.models[1]._f, L0Approximation)
        # `a` needs to be set to something small for the pruning to work as expected
        # pyre-fixme[29]: `Union[(self: TensorBase, indices: Union[None, slice[Any, A...
        self.assertEqual(model_list.models[-1]._f.a, 1e-6)

        # Check transformed objective threshold
        self.assertTrue(
            torch.equal(
                # pyre-fixme[29]: `Union[(self: TensorBase, indices: Union[None, slic...
                acquisition1.acqf.ref_point[-1],
                -self.objective_thresholds_sebo[-1],
            )
        )
        self.assertTrue(
            torch.equal(
                none_throws(acquisition1.objective_thresholds),
                self.objective_thresholds_sebo,
            )
        )
        self.assertEqual(acquisition1.sparsity_threshold, self.X.shape[-1])

        # Check using non-default penalty
        acquisition2 = self.get_acquisition_function(
            options={"penalty": "L1_norm", "target_point": self.target_point},
        )
        self.assertEqual(acquisition2.penalty_name, "L1_norm")
        surrogate = acquisition2.surrogate
        model_list = none_throws(surrogate._model)
        # pyre-fixme[29]: `Union[(self: TensorBase, indices: Union[None, slice[Any, A...
        self.assertIsInstance(model_list.models[1]._f, functools.partial)
        # pyre-fixme[29]: `Union[(self: TensorBase, indices: Union[None, slice[Any, A...
        self.assertIs(model_list.models[1]._f.func, L1_norm_func)

        # assert error raise when constructing non L0/L1 penalty terms
        with self.assertRaisesRegex(
            NotImplementedError, "L2_norm is not currently implemented."
        ):
            self.get_acquisition_function(
                fixed_features=self.fixed_features,
                options={"penalty": "L2_norm", "target_point": self.target_point},
            )

        # assert error raise if target point is not given
        with self.assertRaisesRegex(ValueError, "please provide target point."):
            self.get_acquisition_function(options={"penalty": "L1_norm"})

        # Cache root catches
        with warnings.catch_warnings(record=True) as ws:
            self.get_acquisition_function(
                fixed_features=self.fixed_features,
                options={"target_point": self.target_point},
                botorch_acqf_options={"cache_root": True},
            )
            self.assertEqual(len(ws), 1)
            self.assertEqual(
                "SEBO doesn't support `cache_root=True`. Changing it to `False`.",
                str(ws[0].message),
            )

        # Test with no outcome constraints
        self.get_acquisition_function(
            options={"target_point": self.target_point},
            torch_opt_config=self.torch_opt_config_2,
        )

    @mock.patch(f"{ACQUISITION_PATH}.optimize_acqf")
    def test_optimize_l1(self, mock_optimize_acqf: Mock) -> None:
        mock_optimize_acqf.return_value = (
            torch.tensor([[0.1, 0.1, 0.1], [0.2, 0.2, 0.2]], dtype=torch.double),
            torch.tensor([1.0, 2.0], dtype=torch.double),
        )
        acquisition = self.get_acquisition_function(
            fixed_features=self.fixed_features,
            options={"penalty": "L1_norm", "target_point": self.target_point},
        )
        acquisition.optimize(
            n=2,
            search_space_digest=self.search_space_digest,
            inequality_constraints=self.inequality_constraints,
            fixed_features=self.fixed_features,
            rounding_func=self.rounding_func,
            optimizer_options=self.optimizer_options,
        )

        args, kwargs = mock_optimize_acqf.call_args
        self.assertEqual(kwargs["acq_function"], acquisition.acqf)
        self.assertEqual(kwargs["q"], 2)
        self.assertEqual(kwargs["inequality_constraints"], self.inequality_constraints)
        self.assertEqual(kwargs["post_processing_func"], self.rounding_func)
        self.assertEqual(kwargs["num_restarts"], self.optimizer_options["num_restarts"])
        self.assertEqual(kwargs["raw_samples"], self.optimizer_options["raw_samples"])

    @mock.patch(f"{SEBOACQUISITION_PATH}.optimize_acqf_homotopy")
    def test_optimize_l0(self, mock_optimize_acqf_homotopy: Mock) -> None:
        mock_optimize_acqf_homotopy.return_value = (
            torch.tensor([[0.1, 0.1, 0.1], [0.2, 0.2, 0.2]], dtype=torch.double),
            torch.tensor([1.0, 2.0], dtype=torch.double),
        )
        acquisition = self.get_acquisition_function(
            fixed_features=self.fixed_features,
            options={"penalty": "L0_norm", "target_point": self.target_point},
        )
        acquisition.optimize(
            n=2,
            search_space_digest=self.search_space_digest,
            fixed_features=self.fixed_features,
            rounding_func=self.rounding_func,
            optimizer_options=self.optimizer_options,
        )

        args, kwargs = mock_optimize_acqf_homotopy.call_args
        self.assertEqual(kwargs["acq_function"], acquisition.acqf)
        self.assertEqual(kwargs["q"], 2)
        self.assertEqual(kwargs["post_processing_func"], self.rounding_func)
        self.assertEqual(kwargs["num_restarts"], self.optimizer_options["num_restarts"])
        self.assertEqual(kwargs["raw_samples"], self.optimizer_options["raw_samples"])
        self.assertEqual(
            kwargs["homotopy"]._homotopy_parameters[0].schedule.num_steps,
            self.optimizer_options["num_homotopy_steps"],
        )

        # set self.acqf.cache_pending as False
        acquisition2 = self.get_acquisition_function(
            fixed_features=self.fixed_features,
            options={"penalty": "L0_norm", "target_point": self.target_point},
        )
        # pyre-fixme[16]: `AcquisitionFunction` has no attribute `cache_pending`.
        acquisition2.acqf.cache_pending = torch.tensor(False)
        acquisition2.optimize(
            n=2,
            search_space_digest=self.search_space_digest,
            inequality_constraints=self.inequality_constraints,
            fixed_features=self.fixed_features,
            rounding_func=self.rounding_func,
            optimizer_options=self.optimizer_options,
        )
        _args, kwargs = mock_optimize_acqf_homotopy.call_args
        self.assertEqual(kwargs["acq_function"], acquisition2.acqf)
        self.assertEqual(kwargs["q"], 2)
        self.assertEqual(kwargs["inequality_constraints"], self.inequality_constraints)
        self.assertEqual(kwargs["post_processing_func"], self.rounding_func)
        self.assertEqual(kwargs["num_restarts"], self.optimizer_options["num_restarts"])
        self.assertEqual(kwargs["raw_samples"], self.optimizer_options["raw_samples"])

    def test_optimization_result(self) -> None:
        acquisition = self.get_acquisition_function(
            options={"penalty": "L0_norm", "target_point": self.target_point},
        )
        # Function is f(x) = x_1 on [0, 1]^3, target point of [0.5, 0.5, 0.5].
        # Optimal soln is [1.0, 0.5, 0.5]
        optimizer_options = {Keys.NUM_RESTARTS: 1, Keys.RAW_SAMPLES: 1000}
        xopt, _, _ = acquisition.optimize(
            n=1,
            search_space_digest=self.search_space_digest,
            optimizer_options=optimizer_options,  # pyre-ignore
        )
        self.assertAllClose(xopt[0, 1:], torch.tensor([0.5, 0.5], **self.tkwargs))
        self.assertTrue(
            torch.isclose(xopt[0, 0], torch.tensor(1.0, **self.tkwargs), atol=0.3)
        )

    def test_clamp_to_target(self) -> None:
        X = torch.tensor(
            [[0.5, 0.01, 0.5], [0.05, 0.5, 0.95], [0.1, 0.02, 0.06]], **self.tkwargs
        )
        X_true = torch.tensor(
            [[0.5, 0, 0.5], [0, 0.5, 0.95], [0.1, 0, 0.06]], **self.tkwargs
        )
        self.assertAllClose(
            clamp_to_target(X, torch.zeros(1, 3, **self.tkwargs), 0.05), X_true
        )

    @mock.patch(f"{SEBOACQUISITION_PATH}.optimize_acqf_homotopy")
    @mock.patch(
        f"{SEBOACQUISITION_PATH}.get_batch_initial_conditions",
        wraps=get_batch_initial_conditions,
    )
    def test_get_batch_initial_conditions(
        self, mock_get_batch_initial_conditions: Mock, mock_optimize_acqf_homotopy: Mock
    ) -> None:
        mock_optimize_acqf_homotopy.return_value = (
            torch.tensor([[0.1, 0.1, 0.1], [0.2, 0.2, 0.2]], dtype=torch.double),
            torch.tensor([1.0, 2.0], dtype=torch.double),
        )
        acquisition = self.get_acquisition_function(
            fixed_features={1: 0.5},
            options={"target_point": self.target_point},
            torch_opt_config=self.torch_opt_config_2,
        )
        acquisition.optimize(
            n=2,
            search_space_digest=self.search_space_digest,
            fixed_features=self.fixed_features,
            rounding_func=self.rounding_func,
            optimizer_options={Keys.NUM_RESTARTS: 3, Keys.RAW_SAMPLES: 32},
        )
        call_args = mock_get_batch_initial_conditions.call_args[1]
        self.assertTrue(
            torch.equal(
                call_args["X_pareto"],
                torch.tensor([[0.5, 0.5, 0.5]], dtype=torch.double),
            )
        )
        self.assertTrue(torch.equal(call_args["target_point"], self.target_point))
        self.assertEqual(call_args["raw_samples"], 32)
        self.assertEqual(call_args["num_restarts"], 3)
        # Check the batch initial conditions
        batch_initial_conditions = mock_optimize_acqf_homotopy.call_args[1][
            "batch_initial_conditions"
        ]
        self.assertEqual(batch_initial_conditions.shape, torch.Size([3, 1, 3]))
        self.assertTrue(torch.all(batch_initial_conditions[:1] != 0.5))
        self.assertTrue(torch.all(batch_initial_conditions[1:, :, 1] == 0.5))

    @mock.patch(f"{SEBOACQUISITION_PATH}.optimize_acqf_homotopy")
    @mock.patch(
        f"{SEBOACQUISITION_PATH}.get_batch_initial_conditions",
        wraps=get_batch_initial_conditions,
    )
    def test_optimize_with_provided_batch_initial_conditions(
        self, mock_get_batch_initial_conditions: Mock, mock_optimize_acqf_homotopy: Mock
    ) -> None:
        mock_optimize_acqf_homotopy.return_value = (
            torch.tensor([[0.1, 0.1, 0.1]], dtype=torch.double),
            torch.tensor([1.0], dtype=torch.double),
        )

        # Create batch initial conditions
        batch_ics = torch.rand(3, 1, 3, dtype=torch.double)

        acquisition = self.get_acquisition_function(
            options={
                "target_point": self.target_point,
                "penalty": "L0_norm",
            },
        )

        acquisition.optimize(
            n=1,
            search_space_digest=self.search_space_digest,
            optimizer_options={
                "batch_initial_conditions": batch_ics,
                Keys.NUM_RESTARTS: 3,
                Keys.RAW_SAMPLES: 32,
            },
        )

        # Verify get_batch_initial_conditions was not called
        mock_get_batch_initial_conditions.assert_not_called()

        # Verify the batch_initial_conditions were passed to optimize_acqf_homotopy
        call_kwargs = mock_optimize_acqf_homotopy.call_args[1]
        self.assertTrue(torch.equal(call_kwargs["batch_initial_conditions"], batch_ics))



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


import numpy as np
from ax.utils.common.testutils import TestCase
from ax.utils.common.typeutils import (
    assert_is_instance_dict,
    assert_is_instance_list,
    assert_is_instance_optional,
)
from ax.utils.common.typeutils_nonnative import numpy_type_to_python_type
from pyre_extensions import assert_is_instance


class TestTypeUtils(TestCase):
    def test_assert_is_instance(self) -> None:
        self.assertEqual(assert_is_instance(2.0, float), 2.0)
        with self.assertRaises(TypeError):
            assert_is_instance(2, float)

    def test_assert_is_instance_list(self) -> None:
        self.assertEqual(assert_is_instance_list([1.0, 2.0], float), [1.0, 2.0])
        with self.assertRaises(TypeError):
            assert_is_instance_list([1.0, 2], float)

    def test_assert_is_instance_optional(self) -> None:
        self.assertEqual(assert_is_instance_optional(None, float), None)
        with self.assertRaises(TypeError):
            assert_is_instance_optional(2, float)

    def test_assert_is_instance_dict(self) -> None:
        self.assertEqual(assert_is_instance_dict({"some": 1}, str, int), {"some": 1})
        with self.assertRaises(TypeError):
            assert_is_instance_dict({"some": 1.0}, str, int)
        with self.assertRaises(TypeError):
            assert_is_instance_dict({1: 1}, str, int)

    def test_numpy_type_to_python_type(self) -> None:
        self.assertEqual(type(numpy_type_to_python_type(np.int64(2))), int)
        self.assertEqual(type(numpy_type_to_python_type(np.float64(2))), float)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections.abc import Callable, Generator
from contextlib import contextmanager, ExitStack
from functools import wraps
from typing import Any
from unittest import mock

from botorch.fit import fit_fully_bayesian_model_nuts
from botorch.optim.optimize_mixed import optimize_acqf_mixed_alternating
from botorch.test_utils.mock import mock_optimize_context_manager
from torch import Tensor

try:
    from botorch.utils.multi_objective.optimize import optimize_with_nsgaii

    def minimal_optimize_with_nsgaii(
        *args: Any, **kwargs: Any
    ) -> tuple[Tensor, Tensor]:
        kwargs["population_size"] = 10
        kwargs["max_gen"] = 1
        return optimize_with_nsgaii(*args, **kwargs)
except ImportError:

    def minimal_optimize_with_nsgaii(*args: Any, **kwargs: Any) -> None:
        pass


@contextmanager
def mock_botorch_optimize_context_manager(
    force: bool = False,
) -> Generator[None, None, None]:
    """A context manager that uses mocks to speed up optimization for testing.
    Currently, the primary tactic is to force the underlying scipy methods to
    stop after just one iteration.

    This context manager uses BoTorch's `mock_optimize_context_manager`, and
    adds some additional mocks that are not possible to cover in BoTorch due to
    the need to mock the functions where they are used.

    Args:
        force: If True will not raise an AssertionError if no mocks are called.
            USE RESPONSIBLY.
    """

    def minimal_fit_fully_bayesian(*args: Any, **kwargs: Any) -> None:
        fit_fully_bayesian_model_nuts(*args, **_get_minimal_mcmc_kwargs(**kwargs))

    def minimal_mixed_optimizer(*args: Any, **kwargs: Any) -> tuple[Tensor, Tensor]:
        # BoTorch's `mock_optimize_context_manager` also has some mocks for this,
        # but the full set of mocks applied here cannot be covered by that.
        kwargs["raw_samples"] = 2
        kwargs["num_restarts"] = 1
        kwargs["options"].update(
            {
                "maxiter_alternating": 1,
                "maxiter_continuous": 1,
                "maxiter_init": 1,
                "maxiter_discrete": 1,
            }
        )
        return optimize_acqf_mixed_alternating(*args, **kwargs)

    with ExitStack() as es:
        mock_mcmc_mbm = es.enter_context(
            mock.patch(
                "ax.generators.torch.botorch_modular.utils"
                ".fit_fully_bayesian_model_nuts",
                wraps=minimal_fit_fully_bayesian,
            )
        )

        mock_mixed_optimizer = es.enter_context(
            mock.patch(
                "ax.generators.torch.botorch_modular.acquisition."
                "optimize_acqf_mixed_alternating",
                wraps=minimal_mixed_optimizer,
            )
        )

        mock_nsgaii = es.enter_context(
            mock.patch(
                "ax.generators.torch.botorch_modular.acquisition."
                "optimize_with_nsgaii",
                wraps=minimal_optimize_with_nsgaii,
            )
        )

        es.enter_context(mock_optimize_context_manager())

        yield

        # Only raise if none of the BoTorch or Ax side mocks were called.
        # We do this by catching the error that could be raised by the BoTorch
        # context manager, and combining it with the signals from Ax side mocks.
        try:
            es.close()
        except AssertionError as e:
            # Check if the error is due to no BoTorch mocks being called.
            if "No mocks were called" in str(e):
                botorch_mocks_called = False
            else:
                raise
        else:
            botorch_mocks_called = True

    if (
        not force
        and all(
            mock_.call_count < 1
            for mock_ in [mock_mcmc_mbm, mock_mixed_optimizer, mock_nsgaii]
        )
        and botorch_mocks_called is False
    ):
        raise AssertionError(
            "No mocks were called in the context manager. Please remove unused "
            "mock_botorch_optimize_context_manager()."
        )


def mock_botorch_optimize(f: Callable) -> Callable:
    """Wraps `f` in `mock_botorch_optimize_context_manager` for use as a decorator."""

    @wraps(f)
    # pyre-fixme[3]: Return type must be annotated.
    def inner(*args: Any, **kwargs: Any):
        with mock_botorch_optimize_context_manager():
            return f(*args, **kwargs)

    return inner


@contextmanager
def skip_fit_gpytorch_mll_context_manager() -> Generator[None, None, None]:
    """A context manager that makes `fit_gpytorch_mll` a no-op.

    This should only be used to speed up slow tests.
    """
    with mock.patch(
        "botorch.fit.FitGPyTorchMLL", side_effect=lambda *args, **kwargs: args[0]
    ) as mock_fit:
        yield
    if mock_fit.call_count < 1:
        raise AssertionError(
            "No mocks were called in the context manager. Please remove unused "
            "skip_fit_gpytorch_mll_context_manager()."
        )


def skip_fit_gpytorch_mll(f: Callable) -> Callable:
    """Wraps f in the skip_fit_gpytorch_mll_context_manager for use as a decorator."""

    @wraps(f)
    # pyre-fixme[3]: Return type must be annotated.
    def inner(*args: Any, **kwargs: Any):
        with skip_fit_gpytorch_mll_context_manager():
            return f(*args, **kwargs)

    return inner


def _get_minimal_mcmc_kwargs(**kwargs: Any) -> dict[str, Any]:
    kwargs["warmup_steps"] = 0
    # Just get as many samples as otherwise expected.
    kwargs["num_samples"] = kwargs.get("num_samples", 256) // kwargs.get("thinning", 16)
    kwargs["thinning"] = 1
    return kwargs



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict
from __future__ import annotations

from enum import Enum, unique
from math import ceil, floor
from typing import Any

from ax.core import ObservationFeatures
from ax.core.experiment import Experiment
from ax.core.trial_status import STATUSES_EXPECTING_DATA
from ax.core.utils import get_target_trial_index
from ax.exceptions.generation_strategy import AxGenerationException
from ax.generation_strategy.generation_node import GenerationNode
from ax.utils.common.constants import Keys
from ax.utils.common.func_enum import FuncEnum


@unique
class InputConstructorPurpose(Enum):
    """A simple enum to indicate the purpose of the input constructor.
    Values in this enum will match argument names for ``GenerationNode.gen``.

    Explanation of the different purposes:
        N: Defines the logic to determine the number of arms to generate from the
           next ``GenerationNode`` given the total number of arms expected in
           this trial.
    """

    N = "n"
    FIXED_FEATURES = "fixed_features"

    def __gt__(self, other: InputConstructorPurpose) -> bool:
        """Returns True if the value of this enum member is greater than the value of
        the other enum member. This is used during for sorting any dictionary that
        uses this enum as keys (e.g. ``GenerationNode.input_constructors``).
        """
        return self.value > other.value

    def __lt__(self, other: InputConstructorPurpose) -> bool:
        """Returns True if the value of this enum member is less than the value
        of the other enum member. This is used during for sorting any dictionary that
        use this enum as keys (e.g. ``GenerationNode.input_constructors``).
        """
        return self.value < other.value


class NodeInputConstructors(FuncEnum):
    """An enum which maps to a the name of a callable method for constructing
    ``GenerationNode`` inputs. Recommendation: ends of the names of members
    of this enum should match the corresponding ``InputConstructorPurpose`` name.

    NOTE: All functions defined by this enum should share identical arguments in
    their signatures, and the return type should be the same across all functions
    that are used for the same ``InputConstructorPurpose``.
    """

    ALL_N = "consume_all_n"
    REPEAT_N = "repeat_arm_n"
    REMAINING_N = "remaining_n"
    TARGET_TRIAL_FIXED_FEATURES = "set_target_trial"

    # pyre-ignore[3]: Input constructors will be used to make different inputs,
    # so we need to allow `Any` return type here.
    def __call__(
        self,
        previous_node: GenerationNode | None,
        next_node: GenerationNode,
        gs_gen_call_kwargs: dict[str, Any],
        experiment: Experiment,
    ) -> Any:
        """Defines a method, by which the members of this enum can be called,
        e.g. ``NodeInputConstructors.ALL_N(**kwargs)``, which will call the
        ``consume_all_n`` function from this file, since the name of this
        function corresponds to the value of the enum member ``ALL_N``."""
        return super().__call__(
            previous_node=previous_node,
            next_node=next_node,
            gs_gen_call_kwargs=gs_gen_call_kwargs,
            experiment=experiment,
        )


# ------------------------- Purpose: `fixed_features` ------------------------- #


def set_target_trial(
    previous_node: GenerationNode | None,
    next_node: GenerationNode,
    gs_gen_call_kwargs: dict[str, Any],
    experiment: Experiment,
) -> ObservationFeatures | None:
    """Determine the target trial for the next node based on the current state of the
    ``Experiment``.

     Args:
        previous_node: The previous node in the ``GenerationStrategy``. This is the node
            that is being transition away from, and is provided for easy access to
            properties of this node.
        next_node: The next node in the ``GenerationStrategy``. This is the node that
            will leverage the inputs defined by this input constructor.
        gs_gen_call_kwargs: The kwargs passed to the ``GenerationStrategy``'s
            gen call.
        experiment: The experiment associated with this ``GenerationStrategy``.
    Returns:
        An ``ObservationFeatures`` object that defines the target trial for the next
        node.
    """
    target_trial_idx = get_target_trial_index(experiment=experiment)
    if target_trial_idx is None:
        raise AxGenerationException(
            f"Attempting to construct for input into {next_node} but no trials match "
            "the expected conditions. Often this could be due to no trials on the "
            "experiment that have status quo arm and are in status "
            f"{STATUSES_EXPECTING_DATA} on the experiment. The trials on this "
            f"experiment are: {experiment.trials}."
        )
    return ObservationFeatures(
        parameters={},
        trial_index=target_trial_idx,
    )


# ------------------------- Purpose: `n` ------------------------- #


def consume_all_n(
    previous_node: GenerationNode | None,
    next_node: GenerationNode,
    gs_gen_call_kwargs: dict[str, Any],
    experiment: Experiment,
) -> int:
    """Generate total requested number of arms from the next node.

    Example: Initial exploration with Sobol will generate all arms from a
    single sobol node.

    Note: If no `n` is provided to the ``GenerationStrategy`` gen call, we will use
    the default number of arms for the next node, defined as a constant `DEFAULT_N`
    in the ``GenerationStrategy`` file.

    Args:
        previous_node: The previous node in the ``GenerationStrategy``. This is the node
            that is being transition away from, and is provided for easy access to
            properties of this node.
        next_node: The next node in the ``GenerationStrategy``. This is the node that
            will leverage the inputs defined by this input constructor.
        gs_gen_call_kwargs: The kwargs passed to the ``GenerationStrategy``'s
            gen call.
        experiment: The experiment associated with this ``GenerationStrategy``.
    Returns:
        The total number of requested arms from the next node.
    """
    n_kwarg = gs_gen_call_kwargs.get("n")
    return (
        n_kwarg
        if n_kwarg is not None
        else _get_default_n(experiment=experiment, next_node=next_node)
    )


def repeat_arm_n(
    previous_node: GenerationNode | None,
    next_node: GenerationNode,
    gs_gen_call_kwargs: dict[str, Any],
    experiment: Experiment,
) -> int:
    """Generate a small percentage of arms requested to be used for repeat arms in
    the next trial.

    Note: If no `n` is provided to the ``GenerationStrategy`` gen call, we will use
    the default number of arms for the next node, defined as a constant `DEFAULT_N`
    in the ``GenerationStrategy`` file.

    Args:
        previous_node: The previous node in the ``GenerationStrategy``. This is the node
            that is being transition away from, and is provided for easy access to
            properties of this node.
        next_node: The next node in the ``GenerationStrategy``. This is the node that
            will leverage the inputs defined by this input constructor.
        gs_gen_call_kwargs: The kwargs passed to the ``GenerationStrategy``'s
            gen call.
        experiment: The experiment associated with this ``GenerationStrategy``.
    Returns:
        The number of requested arms from the next node
    """
    n_kwarg = gs_gen_call_kwargs.get("n")
    total_n = (
        n_kwarg
        if n_kwarg is not None
        else _get_default_n(experiment=experiment, next_node=next_node)
    )
    if total_n < 6:
        # if the next trial is small, we don't want to waste allocation on repeat arms
        # users can still manually add repeat arms if they want before allocation
        # and we need to designated this node as skipped for proper transition
        next_node._should_skip = True
        return 0
    elif total_n <= 10:
        return 1
    return ceil(total_n / 10)


def remaining_n(
    previous_node: GenerationNode | None,
    next_node: GenerationNode,
    gs_gen_call_kwargs: dict[str, Any],
    experiment: Experiment,
) -> int:
    """Generate the remaining number of arms requested for this trial in gs.gen().

    Note: If no `n` is provided to the ``GenerationStrategy`` gen call, we will use
    the default number of arms for the next node, defined as a constant `DEFAULT_N`
    in the ``GenerationStrategy`` file.

    Args:
        previous_node: The previous node in the ``GenerationStrategy``. This is the node
            that is being transition away from, and is provided for easy access to
            properties of this node.
        next_node: The next node in the ``GenerationStrategy``. This is the node that
            will leverage the inputs defined by this input constructor.
        gs_gen_call_kwargs: The kwargs passed to the ``GenerationStrategy``'s
            gen call.
        experiment: The experiment associated with this ``GenerationStrategy``.
    Returns:
        The number of requested arms from the next node
    """
    # TODO: @mgarrard improve this logic to be more robust
    grs_this_gen = gs_gen_call_kwargs.get("grs_this_gen", [])
    n_kwarg = gs_gen_call_kwargs.get("n")
    total_n = (
        n_kwarg
        if n_kwarg is not None
        else _get_default_n(experiment=experiment, next_node=next_node)
    )
    # if all arms have been generated, return 0
    return max(total_n - sum(len(gr.arms) for gr in grs_this_gen), 0)


# Helper methods for input constructors
def _get_default_n(experiment: Experiment, next_node: GenerationNode) -> int:
    """Get the default number of arms to generate from the next node.

    Args:
        experiment: The experiment associated with this ``GenerationStrategy``.
        next_node: The next node in the ``GenerationStrategy``. This is the node that
            will leverage the inputs defined by this input constructor.

    Returns:
        The default number of arms to generate from the next node, used if no n is
        provided to the ``GenerationStrategy``'s gen call.
    """
    # If the generator spec contains `n` use that value first
    if next_node.generator_spec_to_gen_from.model_gen_kwargs.get("n") is not None:
        return next_node.generator_spec_to_gen_from.model_gen_kwargs["n"]

    n_for_this_trial = None
    total_concurrent_arms = experiment._properties.get(
        Keys.EXPERIMENT_TOTAL_CONCURRENT_ARMS.value
    )

    # If total_concurrent_arms is set, we will use that in conjunction with the trial
    # type to determine the number of arms to generate from this node
    if total_concurrent_arms is not None:
        if next_node._trial_type == Keys.SHORT_RUN:
            n_for_this_trial = floor(0.5 * total_concurrent_arms)
        elif next_node._trial_type == Keys.LONG_RUN:
            n_for_this_trial = ceil(0.5 * total_concurrent_arms)
        else:
            n_for_this_trial = total_concurrent_arms

    return (
        n_for_this_trial
        if n_for_this_trial is not None
        # GS default n is 1, but these input constructors are used for nodes that
        # should generate more than 1 arm per trial, default to 10
        else next_node.generation_strategy.DEFAULT_N * 10
    )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from abc import ABC
from collections.abc import Iterable

import numpy as np
from ax.adapter.adapter_utils import observed_hypervolume, predicted_hypervolume
from ax.adapter.torch import TorchAdapter
from ax.core.experiment import Experiment
from ax.core.map_data import MAP_KEY, MapData
from ax.core.objective import ScalarizedObjective
from ax.core.optimization_config import (
    MultiObjectiveOptimizationConfig,
    OptimizationConfig,
)
from ax.core.trial import Trial
from ax.core.types import TModelPredictArm, TParameterization
from ax.exceptions.core import UserInputError
from ax.generation_strategy.generation_strategy import GenerationStrategy
from ax.plot.pareto_utils import get_tensor_converter_adapter
from ax.service.utils import best_point as best_point_utils
from ax.service.utils.best_point_utils import select_baseline_name_default_first_trial
from pyre_extensions import assert_is_instance, none_throws


NUM_BINS_PER_TRIAL = 3


class BestPointMixin(ABC):
    experiment: Experiment
    generation_strategy: GenerationStrategy

    def get_best_trial(
        self,
        optimization_config: OptimizationConfig | None = None,
        trial_indices: Iterable[int] | None = None,
        use_model_predictions: bool = True,
    ) -> tuple[int, TParameterization, TModelPredictArm | None] | None:
        """Identifies the best parameterization tried in the experiment so far.

        First attempts to do so with the adapter used in optimization and
        its corresponding predictions if available. Falls back to the best raw
        objective based on the data fetched from the experiment.

        NOTE: ``TModelPredictArm`` is of the form:
            ({metric_name: mean}, {metric_name_1: {metric_name_2: cov_1_2}})

        Args:
            optimization_config: Optimization config to use in place of the one stored
                on the experiment.
            trial_indices: Indices of trials for which to retrieve data. If None will
                retrieve data from all available trials.
            use_model_predictions: Whether to extract the best point using
                model predictions or directly observed values. If ``True``,
                the metric means and covariances in this method's output will
                also be based on model predictions and may differ from the
                observed values.

        Returns:
            Tuple of trial index, parameterization and model predictions for it.
        """
        return self._get_best_trial(
            experiment=self.experiment,
            generation_strategy=self.generation_strategy,
            optimization_config=optimization_config,
            trial_indices=trial_indices,
            use_model_predictions=use_model_predictions,
        )

    def get_best_parameters(
        self,
        optimization_config: OptimizationConfig | None = None,
        trial_indices: Iterable[int] | None = None,
        use_model_predictions: bool = True,
    ) -> tuple[TParameterization, TModelPredictArm | None] | None:
        """Identifies the best parameterization tried in the experiment so far.

        First attempts to do so with the model used in optimization and
        its corresponding predictions if available. Falls back to the best raw
        objective based on the data fetched from the experiment.

        NOTE: ``TModelPredictArm`` is of the form:
            ({metric_name: mean}, {metric_name_1: {metric_name_2: cov_1_2}})

        Args:
            optimization_config: Optimization config to use in place of the one stored
                on the experiment.
            trial_indices: Indices of trials for which to retrieve data. If None will
                retrieve data from all available trials.
            use_model_predictions: Whether to extract the best point using
                model predictions or directly observed values. If ``True``,
                the metric means and covariances in this method's output will
                also be based on model predictions and may differ from the
                observed values.

        Returns:
            Tuple of parameterization and model predictions for it.
        """
        res = self.get_best_trial(
            optimization_config=optimization_config,
            trial_indices=trial_indices,
            use_model_predictions=use_model_predictions,
        )

        if res is None:
            return res

        _, parameterization, vals = res
        return parameterization, vals

    def get_pareto_optimal_parameters(
        self,
        optimization_config: OptimizationConfig | None = None,
        trial_indices: Iterable[int] | None = None,
        use_model_predictions: bool = True,
    ) -> dict[int, tuple[TParameterization, TModelPredictArm]]:
        """Identifies the best parameterizations tried in the experiment so far,
        using model predictions if ``use_model_predictions`` is true and using
        observed values from the experiment otherwise. By default, uses model
        predictions to account for observation noise.

        NOTE: The format of this method's output is as follows:
        { trial_index --> (parameterization, (means, covariances) }, where means
        are a dictionary of form { metric_name --> metric_mean } and covariances
        are a nested dictionary of form
        { one_metric_name --> { another_metric_name: covariance } }.

        Args:
            optimization_config: Optimization config to use in place of the one stored
                on the experiment.
            trial_indices: Indices of trials for which to retrieve data. If None will
                retrieve data from all available trials.
            use_model_predictions: Whether to extract the Pareto frontier using
                model predictions or directly observed values. If ``True``,
                the metric means and covariances in this method's output will
                also be based on model predictions and may differ from the
                observed values.

        Returns:
            A mapping from trial index to the tuple of:
            - the parameterization of the arm in that trial,
            - two-item tuple of metric means dictionary and covariance matrix
                (model-predicted if ``use_model_predictions=True`` and observed
                otherwise).
            Raises a `NotImplementedError` if extracting the Pareto frontier is
            not possible. Note that the returned dict may be empty.
        """
        return self._get_pareto_optimal_parameters(
            experiment=self.experiment,
            generation_strategy=self.generation_strategy,
            optimization_config=optimization_config,
            trial_indices=trial_indices,
            use_model_predictions=use_model_predictions,
        )

    def get_hypervolume(
        self,
        optimization_config: MultiObjectiveOptimizationConfig | None = None,
        trial_indices: Iterable[int] | None = None,
        use_model_predictions: bool = True,
    ) -> float:
        """Calculate hypervolume of a pareto frontier based on either the posterior
        means of given observation features or observed data.

        Args:
            optimization_config: Optimization config to use in place of the one stored
                on the experiment.
            trial_indices: Indices of trials for which to retrieve data. If None will
                retrieve data from all available trials.
            use_model_predictions: Whether to extract the Pareto frontier using
                model predictions or directly observed values. If ``True``,
                the metric means and covariances in this method's output will
                also be based on model predictions and may differ from the
                observed values.
        """
        return self._get_hypervolume(
            experiment=self.experiment,
            generation_strategy=self.generation_strategy,
            optimization_config=optimization_config,
            trial_indices=trial_indices,
            use_model_predictions=use_model_predictions,
        )

    def get_trace_by_progression(
        self,
        optimization_config: OptimizationConfig | None = None,
        bins: list[float] | None = None,
        final_progression_only: bool = False,
    ) -> tuple[list[float], list[float]]:
        """Get the optimization trace with respect to trial progressions instead of
        `trial_indices` (which is the behavior used in `get_trace`). Note that this
        method does not take into account the parallelism of trials and essentially
        assumes that trials are run one after another, in the sense that it considers
        the total number of progressions "used" at the end of trial k to be the
        cumulative progressions "used" in trials 0,...,k. This method assumes that the
        final value of a particular trial is used and does not take the best value
        of a trial over its progressions.

        The best observed value is computed at each value in `bins` (see below for
        details). If `bins` is not supplied, the method defaults to a heuristic of
        approximately `NUM_BINS_PER_TRIAL` per trial, where each trial is assumed to
        run until maximum progression (inferred from the data).

        Args:
            experiment: The experiment to get the trace for.
            optimization_config: An optional optimization config to use for computing
                the trace. This allows computing the traces under different objectives
                or constraints without having to modify the experiment.
            bins: A list progression values at which to calculate the best observed
                value. The best observed value at bins[i] is defined as the value
                observed in trials 0,...,j where j = largest trial such that the total
                progression in trials 0,...,j is less than bins[i].
            final_progression_only: If True, considers the value of the last step to be
                the value of the trial. If False, considers the best along the curve to
                be the value of the trial.

        Returns:
            A tuple containing (1) the list of observed hypervolumes or best values and
            (2) a list of associated x-values (i.e., progressions) useful for plotting.
        """
        return self._get_trace_by_progression(
            experiment=self.experiment,
            optimization_config=optimization_config,
            bins=bins,
            final_progression_only=final_progression_only,
        )

    @staticmethod
    def _get_best_trial(
        experiment: Experiment,
        generation_strategy: GenerationStrategy,
        optimization_config: OptimizationConfig | None = None,
        trial_indices: Iterable[int] | None = None,
        use_model_predictions: bool = True,
    ) -> tuple[int, TParameterization, TModelPredictArm | None] | None:
        optimization_config = optimization_config or none_throws(
            experiment.optimization_config
        )
        if optimization_config.is_moo_problem:
            raise NotImplementedError(
                "Please use `get_pareto_optimal_parameters` for multi-objective "
                "problems."
            )

        if use_model_predictions:
            res = best_point_utils.get_best_parameters_from_model_predictions_with_trial_index(  # noqa
                experiment=experiment,
                adapter=generation_strategy.adapter,
                optimization_config=optimization_config,
                trial_indices=trial_indices,
            )

            if res is not None:
                return res

        return best_point_utils.get_best_by_raw_objective_with_trial_index(
            experiment=experiment,
            optimization_config=optimization_config,
            trial_indices=trial_indices,
        )

    @staticmethod
    def _get_best_observed_value(
        experiment: Experiment,
        optimization_config: OptimizationConfig | None = None,
        trial_indices: Iterable[int] | None = None,
    ) -> float | None:
        """Identifies the best objective value observed in the experiment
        among the trials indicated by `trial_indices`.

        Args:
            experiment: The experiment to get the best objective value for.
            optimization_config: Optimization config to use in place of the one stored
                on the experiment.
            trial_indices: Indices of trials for which to retrieve data. If None will
                retrieve data from all available trials.

        Returns:
            The best objective value so far.
        """
        if optimization_config is None:
            optimization_config = none_throws(experiment.optimization_config)
        if optimization_config.is_moo_problem:
            raise NotImplementedError(
                "Please use `get_hypervolume` for multi-objective problems."
            )

        res = best_point_utils.get_best_by_raw_objective_with_trial_index(
            experiment=experiment,
            optimization_config=optimization_config,
            trial_indices=trial_indices,
        )

        predictions = res[2] if res is not None else None
        if predictions is None:
            return None

        means = none_throws(predictions)[0]
        objective = optimization_config.objective
        if isinstance(objective, ScalarizedObjective):
            value = 0
            for metric, weight in objective.metric_weights:
                value += means[metric.name] * weight
            return value
        else:
            name = objective.metric_names[0]
            return means[name]

    @staticmethod
    def _get_pareto_optimal_parameters(
        experiment: Experiment,
        generation_strategy: GenerationStrategy,
        optimization_config: OptimizationConfig | None = None,
        trial_indices: Iterable[int] | None = None,
        use_model_predictions: bool = True,
    ) -> dict[int, tuple[TParameterization, TModelPredictArm]]:
        optimization_config = optimization_config or none_throws(
            experiment.optimization_config
        )
        if not optimization_config.is_moo_problem:
            raise NotImplementedError(
                "Please use `get_best_parameters` for single-objective problems."
            )
        return best_point_utils.get_pareto_optimal_parameters(
            experiment=experiment,
            generation_strategy=generation_strategy,
            optimization_config=optimization_config,
            trial_indices=trial_indices,
            use_model_predictions=use_model_predictions,
        )

    @staticmethod
    def _get_hypervolume(
        experiment: Experiment,
        generation_strategy: GenerationStrategy,
        optimization_config: MultiObjectiveOptimizationConfig | None = None,
        trial_indices: Iterable[int] | None = None,
        use_model_predictions: bool = True,
    ) -> float:
        data = experiment.lookup_data()
        if len(data.df) == 0:
            return 0.0
        moo_optimization_config = assert_is_instance(
            optimization_config or experiment.optimization_config,
            MultiObjectiveOptimizationConfig,
        )

        if use_model_predictions:
            # Make sure that the adapter is fitted. If adapter is fitted already,
            # this should be a no-op.
            generation_strategy._curr._fit(experiment=experiment)
            adapter = generation_strategy.adapter
            if not isinstance(adapter, TorchAdapter):
                raise ValueError(
                    f"Adapter {adapter} is not of type TorchAdapter, cannot "
                    "calculate predicted hypervolume."
                )
            return predicted_hypervolume(
                adapter=adapter, optimization_config=optimization_config
            )

        minimal_adapter = get_tensor_converter_adapter(
            experiment=experiment,
            data=experiment.lookup_data(trial_indices=trial_indices),
        )

        return observed_hypervolume(
            adapter=minimal_adapter, optimization_config=moo_optimization_config
        )

    @staticmethod
    def _get_trace_by_progression(
        experiment: Experiment,
        optimization_config: OptimizationConfig | None = None,
        bins: list[float] | None = None,
        final_progression_only: bool = False,
    ) -> tuple[list[float], list[float]]:
        optimization_config = optimization_config or none_throws(
            experiment.optimization_config
        )
        objective = optimization_config.objective.metric.name
        minimize = optimization_config.objective.minimize
        map_data = experiment.lookup_data()
        if not isinstance(map_data, MapData):
            raise ValueError("`get_trace_by_progression` requires MapData.")
        map_df = map_data.map_df

        map_df = map_df[map_df["metric_name"] == objective]
        map_df = map_df.sort_values(by=["trial_index", MAP_KEY])
        df = (
            map_df.drop_duplicates(MapData.DEDUPLICATE_BY_COLUMNS, keep="last")
            if final_progression_only
            else map_df
        )

        # compute cumulative steps
        prev_steps_df = map_df.drop_duplicates(
            MapData.DEDUPLICATE_BY_COLUMNS, keep="last"
        )[["trial_index", MAP_KEY]].copy()

        # shift the cumsum by one so that we count cumulative steps not including
        # the current trial
        prev_steps_df[MAP_KEY] = (
            prev_steps_df[MAP_KEY].cumsum().shift(periods=1).fillna(0)
        )
        prev_steps_df = prev_steps_df.rename(columns={MAP_KEY: "prev_steps"})
        df = df.merge(prev_steps_df, on=["trial_index"])
        df["cumulative_steps"] = df[MAP_KEY] + df["prev_steps"]
        progressions = df["cumulative_steps"].to_numpy()

        if bins is None:
            # this assumes that there is at least one completed trial that
            # reached the maximum progression
            prog_per_trial = df[MAP_KEY].max()
            num_trials = len(experiment.trials)
            bins = np.linspace(
                0, prog_per_trial * num_trials, NUM_BINS_PER_TRIAL * num_trials
            )
        else:
            bins = np.array(bins)  # pyre-ignore[9]

        # pyre-fixme[9]: bins has type `Optional[List[float]]`; used as
        #  `ndarray[typing.Any, dtype[typing.Any]]`.
        bins = np.expand_dims(bins, axis=0)

        # compute for each bin value the largest trial index finished by then
        # (interpreting the bin value as a cumulative progression)
        best_observed_idcs = np.maximum.accumulate(
            np.argmax(np.expand_dims(progressions, axis=1) >= bins, axis=0)
        )
        obj_vals = (df["mean"].cummin() if minimize else df["mean"].cummax()).to_numpy()
        best_observed = obj_vals[best_observed_idcs]
        # pyre-fixme[16]: Item `List` of `Union[List[float], ndarray[typing.Any,
        #  np.dtype[typing.Any]]]` has no attribute `squeeze`.
        return best_observed.tolist(), bins.squeeze(axis=0).tolist()

    def get_improvement_over_baseline(
        self,
        experiment: Experiment,
        generation_strategy: GenerationStrategy,
        baseline_arm_name: str | None = None,
    ) -> float:
        """Returns the scalarized improvement over baseline, if applicable.

        Returns:
            For Single Objective cases, returns % improvement of objective.
            Positive indicates improvement over baseline. Negative indicates regression.
            For Multi Objective cases, throws NotImplementedError
        """
        if experiment.is_moo_problem:
            raise NotImplementedError(
                "`get_improvement_over_baseline` not yet implemented"
                + " for multi-objective problems."
            )
        if not baseline_arm_name:
            baseline_arm_name, _ = select_baseline_name_default_first_trial(
                experiment=experiment,
                baseline_arm_name=baseline_arm_name,
            )

        optimization_config = experiment.optimization_config
        if not optimization_config:
            raise ValueError("No optimization config found.")

        objective_metric_name = optimization_config.objective.metric.name

        # get the baseline trial
        data = experiment.lookup_data().df
        data = data[data["arm_name"] == baseline_arm_name]
        if len(data) == 0:
            raise UserInputError(
                "`get_improvement_over_baseline`"
                " could not find baseline arm"
                f" `{baseline_arm_name}` in the experiment data."
            )
        data = data[data["metric_name"] == objective_metric_name]
        baseline_value = data.iloc[0]["mean"]

        # Find objective value of the best trial
        idx, param, best_arm = none_throws(
            self._get_best_trial(
                experiment=experiment,
                generation_strategy=generation_strategy,
                optimization_config=optimization_config,
                use_model_predictions=False,
            )
        )
        best_arm = none_throws(best_arm)
        best_obj_value = best_arm[0][objective_metric_name]

        def percent_change(x: float, y: float, minimize: bool) -> float:
            if x == 0:
                raise ZeroDivisionError(
                    "Cannot compute percent improvement when denom is zero"
                )
            percent_change = (y - x) / abs(x) * 100
            if minimize:
                percent_change = -percent_change
            return percent_change

        return percent_change(
            x=baseline_value,
            y=best_obj_value,
            minimize=optimization_config.objective.minimize,
        )

    @staticmethod
    def _to_best_point_tuple(
        experiment: Experiment,
        trial_index: int,
        parameterization: TParameterization,
        model_prediction: TModelPredictArm | None,
    ) -> tuple[TParameterization, dict[str, float | tuple[float, float]], int, str]:
        """
        Return the tuple expected by the return signature of get_best_parameterization
        and get_pareto_frontier in the Ax API.

        TODO: Remove this helper when we clean up BestPointMixin.

        Returns:
            - The parameters predicted to have the best optimization value without
                violating any outcome constraints.
            - The metric values for the best parameterization. Uses model prediction if
                use_model_predictions=True, otherwise returns observed data.
            - The trial which most recently ran the best parameterization
            - The name of the best arm (each trial has a unique name associated with
                each parameterization)
        """

        if model_prediction is not None:
            mean, covariance = model_prediction

            prediction: dict[str, float | tuple[float, float]] = {
                metric_name: (
                    mean[metric_name],
                    none_throws(covariance)[metric_name][metric_name],
                )
                for metric_name in mean.keys()
            }
        else:
            data_dict = experiment.lookup_data(trial_indices=[trial_index]).df.to_dict()

            prediction: dict[str, float | tuple[float, float]] = {
                data_dict["metric_name"][i]: (data_dict["mean"][i], data_dict["sem"][i])
                for i in range(len(data_dict["metric_name"]))
            }

        trial = assert_is_instance(experiment.trials[trial_index], Trial)
        arm = none_throws(trial.arm)

        return parameterization, prediction, trial_index, arm.name



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from ax.adapter.transforms.base import Transform
from ax.adapter.transforms.bilog_y import BilogY
from ax.adapter.transforms.choice_encode import (
    ChoiceEncode,
    ChoiceToNumericChoice,
    OrderedChoiceEncode,
    OrderedChoiceToIntegerRange,
)
from ax.adapter.transforms.convert_metric_names import ConvertMetricNames
from ax.adapter.transforms.derelativize import Derelativize
from ax.adapter.transforms.fill_missing_parameters import FillMissingParameters
from ax.adapter.transforms.fixed_to_tunable import FixedToTunable
from ax.adapter.transforms.int_range_to_choice import IntRangeToChoice
from ax.adapter.transforms.int_to_float import IntToFloat, LogIntToFloat
from ax.adapter.transforms.ivw import IVW
from ax.adapter.transforms.log import Log
from ax.adapter.transforms.log_y import LogY
from ax.adapter.transforms.logit import Logit
from ax.adapter.transforms.map_key_to_float import MapKeyToFloat
from ax.adapter.transforms.merge_repeated_measurements import MergeRepeatedMeasurements
from ax.adapter.transforms.metadata_to_float import MetadataToFloat
from ax.adapter.transforms.metadata_to_task import MetadataToTask
from ax.adapter.transforms.metrics_as_task import MetricsAsTask
from ax.adapter.transforms.one_hot import OneHot
from ax.adapter.transforms.power_transform_y import PowerTransformY
from ax.adapter.transforms.relativize import Relativize, RelativizeWithConstantControl
from ax.adapter.transforms.remove_fixed import RemoveFixed
from ax.adapter.transforms.search_space_to_choice import SearchSpaceToChoice
from ax.adapter.transforms.standardize_y import StandardizeY
from ax.adapter.transforms.stratified_standardize_y import StratifiedStandardizeY
from ax.adapter.transforms.task_encode import TaskChoiceToIntTaskChoice, TaskEncode
from ax.adapter.transforms.time_as_feature import TimeAsFeature
from ax.adapter.transforms.transform_to_new_sq import TransformToNewSQ
from ax.adapter.transforms.trial_as_task import TrialAsTask
from ax.adapter.transforms.unit_x import UnitX
from ax.adapter.transforms.winsorize import Winsorize


"""
A registry of transform classes for storage.

Transforms are stored in the DB as part of the JSON-encoded
GenerationNode (or Step) objects. When loading the GenerationStrategy,
we will look up the transform class that matches the stringified
transform type in JSON object, and return the matching transform class.

NOTE: If removing a transform, please add it to REMOVED_TRANSFORMS.
These will be discarded while loading experiments from the DB.
If deprecating a transform in favor of a new one (i.e. renaming),
please add it to DEPRECATED_TRANSFORMS. When loading from the DB,
we will return the replacement class.
"""
TRANSFORM_REGISTRY: set[type[Transform]] = {
    ConvertMetricNames,
    Derelativize,
    FixedToTunable,
    IntRangeToChoice,
    IntToFloat,
    IVW,
    Log,
    OneHot,
    OrderedChoiceEncode,  # TO BE DEPRECATED
    OrderedChoiceToIntegerRange,
    # This transform was upstreamed into the base adapter.
    # DEPRECATED: OutOfDesign
    RemoveFixed,
    SearchSpaceToChoice,
    StandardizeY,
    StratifiedStandardizeY,
    TaskEncode,  # TO BE DEPRECATED
    TaskChoiceToIntTaskChoice,
    TrialAsTask,
    UnitX,
    Winsorize,
    # CapParameter,  DEPRECATED
    PowerTransformY,
    ChoiceEncode,  # TO BE DEPRECATED
    ChoiceToNumericChoice,
    Logit,
    # MapUnitX, DEPRECATED
    MetadataToFloat,
    MetadataToTask,
    MetricsAsTask,
    LogY,
    Relativize,
    RelativizeWithConstantControl,
    MergeRepeatedMeasurements,
    TimeAsFeature,
    TransformToNewSQ,
    FillMissingParameters,
    LogIntToFloat,
    MapKeyToFloat,
    BilogY,
}

REMOVED_TRANSFORMS: set[str] = {
    "OutOfDesign",
    "CapParameter",
    "MapUnitX",
}

"""
Dict mapping transforms which are being deprecated to the transforms replacing them.
These will be present in TRANSFORM_REGISTRY so that old call sites
can still store properly, but when loading back the new class will be used.
"""
DEPRECATED_TRANSFORMS: dict[str, type[Transform]] = {
    "OrderedChoiceEncode": OrderedChoiceToIntegerRange,
    "ChoiceEncode": ChoiceToNumericChoice,
    "TaskEncode": TaskChoiceToIntTaskChoice,
}

REVERSE_TRANSFORM_REGISTRY: dict[str, type[Transform]] = {
    k.__name__: k for k in TRANSFORM_REGISTRY if k not in DEPRECATED_TRANSFORMS
}


def register_transform(transform: type[Transform]) -> None:
    """Register a transform class for storage.

    This is intended for external additions. Transforms that are available in
    the core library should be added to the TRANSFORM_REGISTRY directly.
    """
    TRANSFORM_REGISTRY.add(transform)
    REVERSE_TRANSFORM_REGISTRY[transform.__name__] = transform



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from warnings import warn

import numpy as np

from ax.benchmark.benchmark_problem import BenchmarkProblem, get_soo_opt_config
from ax.benchmark.benchmark_test_functions.synthetic import IdentityTestFunction
from ax.core.parameter import ChoiceParameter, ParameterType
from ax.core.search_space import SearchSpace


def get_baseline(num_choices: int, n_sims: int = 100000000) -> float:
    """
    Compute the baseline value.

    The baseline for this problem takes into account noise, because it uses the
    inference trace, and the bandit structure, which allows for running all arms
    in one noisy batch:

    Run a BatchTrial with every arm, with equal size.  Choose the arm with the
    best observed value and take its true value.  Take the expectation of the
    outcome of this process.
    """
    noise_per_arm = num_choices**0.5
    sim_observed_effects = (
        np.random.normal(0, noise_per_arm, (n_sims, num_choices))
        + np.arange(num_choices)[None, :]
    )
    identified_best_arm = sim_observed_effects.argmin(axis=1)
    # because of the use of IdentityTestFunction
    baseline = identified_best_arm.mean()
    return baseline


def get_bandit_problem(num_choices: int = 30, num_trials: int = 3) -> BenchmarkProblem:
    parameter = ChoiceParameter(
        name="x0",
        parameter_type=ParameterType.INT,
        values=list(range(num_choices)),
        is_ordered=False,
        sort_values=False,
    )
    search_space = SearchSpace(parameters=[parameter])
    test_function = IdentityTestFunction()
    optimization_config = get_soo_opt_config(
        outcome_names=test_function.outcome_names, observe_noise_sd=True
    )
    baselines = {
        10: 1.40736478,
        30: 2.4716703,
        100: 4.403284,
    }
    if num_choices not in baselines:
        warn(
            f"Baseline value is not available for num_choices={num_choices}. Use "
            "`get_baseline` to compute the baseline and add it to `baselines`."
        )
        baseline_value = baselines[30]
    else:
        baseline_value = baselines[num_choices]
    return BenchmarkProblem(
        name="Bandit",
        num_trials=num_trials,
        search_space=search_space,
        optimization_config=optimization_config,
        optimal_value=0,
        baseline_value=baseline_value,
        test_function=test_function,
        report_inference_value_as_trace=True,
        noise_std=1.0,
        status_quo_params={"x0": num_choices // 2},
    )
